# coding=utf-8
"""
    数据对外接口模块，其它模块需要数据都只应该使用SymbolPd, 不应涉及其它内部模块的使用
"""
from collections import Iterable
import pandas as pd
import numpy as np
from ultron.ump.core import env
from ultron.ump.core.fixes import six
from ultron.ump.market.market import split_k_market
from ultron.kdutils.date import week_of_date


def _benchmark(df, benchmark_pd):
    """
    在内部使用kline_pd获取金融时间序列pd.DataFrame后，如果参数中
    基准benchmark（pd.DataFrame对象）存在，使用基准benchmark的
    时间范围切割kline_pd返回的金融时间序列
    :param df: 金融时间序列pd.DataFrame对象
    :param benchmark: 资金回测时间标尺，AbuBenchmark实例对象
    :return: 使用基准的时间范围切割返回的金融时间序列
    """
    if len(df.index & benchmark_pd.index) <= 0:
        # 如果基准benchmark时间范围和输入的df没有交集，直接返回None
        return None

    df = df.reindex(benchmark_pd.index)
    kl_pd = df.loc[benchmark_pd.index]

    # 两个金融时间序列通过loc寻找交集
    kl_pd = df.loc[benchmark_pd.index]
    # nan的date个数即为不相交的个数
    nan_cnt = kl_pd['date'].isnull().value_counts()
    # 两个金融序列是否相同的结束日期
    same_end = df.index[-1] == benchmark_pd.index[-1]
    # 两个金融序列是否相同的开始日期
    same_head = df.index[0] == benchmark_pd.index[0]

    # 如果nan_cnt即不相交个数大于benchmark基准个数的1/3，放弃
    base_keep_div = 3
    if same_end or same_head:
        # 如果两个序列有相同的开始或者结束改为1/2，也就是如果数据头尾日起的标尺有一个对的上的话，放宽na数量
        base_keep_div = 2
    if same_end and same_head:
        # 如果两个序列同时有相同的开始和结束改为1，也就是如果数据头尾日起的标尺都对的上的话，na数量忽略不计
        base_keep_div = 1

    #if symbol.is_a_stock():
    #    # 如果是A股市场的目标，由于停盘频率和周期都会长与其它市场所以再放宽一些
    #    base_keep_div *= 0.7

    if nan_cnt.index.shape[0] > 0 and nan_cnt.index.tolist().count(True) > 0 \
            and nan_cnt.loc[True] > benchmark_pd.shape[0] / base_keep_div:
        # nan 个数 > 基准base_keep_div分之一放弃
        return None

    # 来到这里说明没有放弃，那么就填充nan
    # 首先nan的交易量是0
    kl_pd.volume.fillna(value=0, inplace=True)
    # nan的p_change是0
    kl_pd.p_change.fillna(value=0, inplace=True)
    # 先把close填充了，然后用close填充其它的
    kl_pd.close.fillna(method='pad', inplace=True)
    kl_pd.close.fillna(method='bfill', inplace=True)
    # 用close填充open
    kl_pd.open.fillna(value=kl_pd.close, inplace=True)
    # 用close填充high
    kl_pd.high.fillna(value=kl_pd.close, inplace=True)
    # 用close填充low
    kl_pd.low.fillna(value=kl_pd.close, inplace=True)
    # 用close填充pre_close
    kl_pd.pre_close.fillna(value=kl_pd.close, inplace=True)

    # 细节nan处理完成后，把剩下的nan都填充了
    kl_pd = kl_pd.fillna(method='pad')
    # bfill再来一遍只是为了填充最前面的nan
    kl_pd.fillna(method='bfill', inplace=True)

    # pad了数据所以，交易日期date的值需要根据time index重新来一遍
    kl_pd['date'] = [int(ts.date().strftime("%Y%m%d")) for ts in kl_pd.index]
    kl_pd['date_week'] = kl_pd['date'].apply(
        lambda x: week_of_date(str(x), '%Y%m%d'))

    return kl_pd


# noinspection PyDeprecation
def make_kl_df(symbol,
               data_mode=env.EMarketDataSplitMode.E_DATA_SPLIT_SE,
               n_folds=2,
               start=None,
               end=None,
               benchmark=None,
               show_progress=True,
               parallel=False,
               parallel_save=True):
    """
    外部获取金融时间序列接口
    eg: n_fold=2, start=None, end=None ，从今天起往前数两年
        n_fold=2, start='2015-02-14', end=None， 从2015-02-14到现在，n_fold无效
        n_fold=2, start=None, end='2016-02-14'，从2016-02-14起往前数两年
        n_fold=2, start='2015-02-14', end='2016-02-14'，从start到end

    :param data_mode: EMarketDataSplitMode对象
    :param symbol: list or Series or str or Symbol
                    e.g :['TSLA','SFUN'] or 'TSLA' or Symbol(MType.US,'TSLA')
    :param n_folds: 请求几年的历史回测数据int
    :param start: 请求的开始日期 str对象
    :param end: 请求的结束日期 str对象
    :param benchmark: 资金回测时间标尺，AbuBenchmark实例对象
    :param show_progress: 是否显示进度条
    :param parallel: 是否并行获取
    :param parallel_save: 是否并行后进行统一批量保存
    """

    if isinstance(symbol, (list, tuple, pd.Series, pd.Index)):
        # 如果symbol是可迭代的序列对象，最终返回三维面板数据pd.Panel
        panel = dict()
        if parallel:
            # 如果并行获取
            if env.g_data_fetch_mode != env.EMarketDataFetchMode.E_DATA_FETCH_FORCE_NET \
                    and env.g_data_cache_type == env.EDataCacheType.E_DATA_CACHE_HDF5:
                # 只能针对非hdf5存贮形式下或者针对force_net，因为hdf5多线程读容易卡死
                raise RuntimeError(
                    'make_kl_df just suit force net or not hdf5 store!')

            df_dicts = kl_df_dict_parallel(symbol,
                                           data_mode=data_mode,
                                           n_folds=n_folds,
                                           start=start,
                                           end=end,
                                           benchmark=benchmark,
                                           save=parallel_save,
                                           how='thread')
            for df_dict in df_dicts:
                for key_tuple, df in df_dict.values():
                    if df is None or df.shape[0] == 0:
                        continue
                    # 即丢弃原始df_dict保存金融时间序列时使用的save_kl_key，只保留df，赋予panel
                    panel[key_tuple[0].value] = df
        else:

            def _batch_make_kl_df():
                with AbuMulPidProgress(len(symbol),
                                       '_make_kl_df complete') as progress:
                    for pos, _symbol in enumerate(symbol):
                        _df, _ = _make_kl_df(_symbol,
                                             data_mode=data_mode,
                                             n_folds=n_folds,
                                             start=start,
                                             end=end,
                                             benchmark=benchmark,
                                             save=True)
                        if show_progress:
                            progress.show()
                        # TODO 做pd.Panel数据应该保证每一个元素的行数和列数都相等，不是简单的有数据就行
                        if _df is None or _df.shape[0] == 0:
                            continue

                        panel[symbol[pos]] = _df

            _batch_make_kl_df()
        # TODO pd.Panel过时
        return pd.Panel(panel)

    elif isinstance(symbol, Symbol) or isinstance(symbol, six.string_types):
        # 对单个symbol进行数据获取
        df, _ = _make_kl_df(symbol,
                            data_mode=data_mode,
                            n_folds=n_folds,
                            start=start,
                            end=end,
                            benchmark=benchmark,
                            save=True)
        return df
    else:
        raise TypeError('symbol type is error')


def kl_df_dict_parallel(symbols,
                        data_mode=env.EMarketDataSplitMode.E_DATA_SPLIT_SE,
                        n_folds=2,
                        start=None,
                        end=None,
                        benchmark=None,
                        n_jobs=16,
                        how='thread'):
    """
    多进程或者多线程对外执行函数，多任务批量获取时间序列数据
    :param symbols: symbol序列
    :param data_mode: EMarketDataSplitMode enum对象
    :param n_folds: 请求几年的历史回测数据int
    :param start: 请求的开始日期 str对象
    :param end: 请求的结束日期 str对象
    :param benchmark: 资金回测时间标尺，AbuBenchmark实例对象
    :param n_jobs: 并行的任务数，对于进程代表进程数，线程代表线程数
    :param save: 是否统一进行批量保存，即在批量获取金融时间序列后，统一进行批量保存，默认True
    :param how: process：多进程，thread：多线程，main：单进程单线程
    """

    # TODO Iterable和six.string_types的判断抽出来放在一个模块，做为Iterable的判断来使用
    if not isinstance(symbols, Iterable) or isinstance(symbols,
                                                       six.string_types):
        # symbols必须是可迭代的序列对象
        raise TypeError('symbols must a Iterable obj!')
    # 可迭代的symbols序列分成n_jobs个子序列
    parallel_symbols = split_k_market(n_jobs, market_symbols=symbols)
    # 使用partial对并行函数_kl_df_dict_parallel进行委托
    parallel_func = partial(_kl_df_dict_parallel,
                            data_mode=data_mode,
                            n_folds=n_folds,
                            start=start,
                            end=end,
                            benchmark=benchmark)
    # 因为切割会有余数，所以将原始设置的进程数切换为分割好的个数, 即32 -> 33 16 -> 17
    n_jobs = len(parallel_symbols)
    if how == 'process':
        """
            mac os 10.9 以后的并行加上numpy不是crash就是进程卡死，不要用，用thread
        """
        if ABuEnv.g_is_mac_os:
            logging.info('mac os 10.9 parallel with numpy crash or dead!!')

        parallel = Parallel(n_jobs=n_jobs, verbose=0, pre_dispatch='2*n_jobs')
        df_dicts = parallel(
            delayed(parallel_func)(choice_symbols)
            for choice_symbols in parallel_symbols)
    elif how == 'thread':
        # 通过ThreadPoolExecutor进行线程并行任务
        with ThreadPoolExecutor(max_workers=n_jobs) as pool:
            k_use_map = True
            if k_use_map:
                df_dicts = list(pool.map(parallel_func, parallel_symbols))
            else:
                futures = [
                    pool.submit(parallel_func, symbols)
                    for symbols in parallel_symbols
                ]
                df_dicts = [
                    future.result() for future in futures
                    if future.exception() is None
                ]
    elif how == 'main':
        # 单进程单线程
        df_dicts = [parallel_func(symbols) for symbols in parallel_symbols]
    else:
        raise TypeError('ONLY process OR thread!')
    return df_dicts


# -*- coding: utf-8 -*-

from ultron.tradingday.Enums._TimeUnits cimport TimeUnits as tu

cpdef enum TimeUnits:
    BDays = tu.BDays
    Days = tu.Days
    Weeks = tu.Weeks
    Months = tu.Months
    Years = tu.Years


APPROX_BDAYS_PER_MONTH = 21
APPROX_BDAYS_PER_YEAR = 252

MONTHS_PER_YEAR = 12
WEEKS_PER_YEAR = 52
QTRS_PER_YEAR = 4

DAILY = 'daily'
WEEKLY = 'weekly'
MONTHLY = 'monthly'
QUARTERLY = 'quarterly'
YEARLY = 'yearly'

ANNUALIZATION_FACTORS = {
    DAILY: APPROX_BDAYS_PER_YEAR,
    WEEKLY: WEEKS_PER_YEAR,
    MONTHLY: MONTHS_PER_YEAR,
    QUARTERLY: QTRS_PER_YEAR,
    YEARLY: 1
}


import os, time
import datetime as dt
import sqlalchemy
import numpy as np
import pandas as pd
import statsmodels.api as sm


class LongShortWeighted(object):

    def __init__(self):
        self._industry_styles = [
            'Bank', 'RealEstate', 'Health', 'Transportation', 'Mining',
            'NonFerMetal', 'HouseApp', 'LeiService', 'MachiEquip', 'BuildDeco',
            'CommeTrade', 'CONMAT', 'Auto', 'Textile', 'FoodBever',
            'Electronics', 'Computer', 'LightIndus', 'Utilities', 'Telecom',
            'AgriForest', 'CHEM', 'Media', 'IronSteel', 'NonBankFinan',
            'ELECEQP', 'AERODEF', 'Conglomerates'
        ]
        self._risk_styles = [
            'BETA', 'MOMENTUM', 'SIZE', 'EARNYILD', 'RESVOL', 'GROWTH', 'BTOP',
            'LEVERAGE', 'LIQUIDTY', 'SIZENL'
        ]

    #暂时写此处，去极值
    def winsorize(self,
                  total_data,
                  method='sigma',
                  limits=(3.0, 3.0),
                  drop=True):
        se = total_data.copy()
        if method == 'quantile':
            down, up = se.quantile([limits[0], 1.0 - limits[1]])
        elif method == 'sigma':
            std, mean = se.std(), se.mean()
            down, up = mean - limits[0] * std, mean + limits[1] * std

        if drop:
            se[se < down] = np.NaN
            se[se > up] = np.NaN
        else:
            se[se < down] = down
            se[se > up] = up
        return se

    #标准化
    def standardize(self, total_data):
        try:
            res = (total_data - total_data.mean()) / total_data.std()
        except:
            res = pd.Series(data=np.NaN, index=total_data.index)
        return res

    #中性化
    def neutralize(self, total_data, risk_df):
        se = total_data.dropna()
        # se = total_data.copy()
        risk = risk_df.loc[se.index, :]
        # use numpy for neu, which is faster
        x = np.linalg.lstsq(risk.values, np.matrix(se).T)[0]
        se_neu = se - risk.dot(x)[0]

        return se_neu

    def _to_weights(self, group):
        demeaned_vals = group - group.mean()
        return demeaned_vals / demeaned_vals.abs().sum()

    def _to_ls_count(self, group, long=True):
        demeaned_vals = group - group.mean()
        if long:
            count = len(demeaned_vals[demeaned_vals > 0])
        else:
            count = len(demeaned_vals[demeaned_vals < 0])
        return count

    def returns(self, total_data, forward_returns, init_capital=100000):
        weights = total_data.groupby(level=['trade_date']).apply(
            self._to_weights)
        weighted_returns = forward_returns.multiply(weights, axis=0)

        factor_ret = weighted_returns.groupby(level='trade_date').sum()
        pnl = init_capital * 2 * factor_ret
        turnover = weights.unstack().T.diff().abs().sum(axis=1)
        long_count = total_data.groupby(level=['trade_date']).apply(
            self._to_ls_count)
        short_count = total_data.groupby(level=['trade_date']).apply(
            lambda x: self._to_ls_count(x, long=False))
        return factor_ret, pnl, turnover, long_count, short_count

    def run(self,
            factor_data,
            risk_data,
            forward_returns,
            factor_name,
            horizon=1,
            method='quantile',
            up_limit=0.025,
            down_limit=0.025,
            init_capital=100000):
        """
        参数：
            horizon: 调仓期，按照交易日计算。
        """
        factor_se = factor_data.set_index(['code', 'trade_date'])
        risk_se = risk_data.set_index(['code', 'trade_date'])
        forward_returns = forward_returns.set_index(['code', 'trade_date'])
        # risk_df = risk_se.reindex(factor_se.index)[self._industry_styles + ['SIZE'] + ['COUNTRY']]
        risk_df = risk_se.reindex(factor_se.index)[self._industry_styles +
                                                   ['SIZE']]
        risk_df.dropna(inplace=True)
        factor_se = factor_se.loc[risk_df.index][factor_name]
        returns = forward_returns.loc[risk_df.index]

        factor_se = factor_se.groupby(
            level='trade_date').apply(lambda x: self.winsorize(
                x, method=method, limits=(up_limit, down_limit)))
        # factor_se = factor_se.groupby(level='trade_date').apply(lambda x: self.neutralize(x, risk_df)) # index 多了一项
        grouped = factor_se.groupby(level='trade_date')
        res = []
        for trade_date, g in grouped:
            neu = self.neutralize(g, risk_df)
            res.append(neu)
        factor_se = pd.concat(res, axis=0)

        factor_se = factor_se.groupby(
            level='trade_date').apply(lambda x: self.standardize(x))

        factor_ret, pnl, turnover_se, lc, sc = self.returns(
            factor_se, forward_returns['ret'], init_capital=init_capital)

        #计算指标
        ir = factor_ret.mean() / factor_ret.std()
        sharpe = np.sqrt(252 / horizon) * ir
        turnover = turnover_se.mean()
        returns = factor_ret.sum() * 252 / horizon / len(factor_ret)
        fitness = sharpe * np.sqrt(abs(returns) / turnover)
        margin = factor_ret.sum() / turnover_se.sum()

        capital = pnl + init_capital
        running_max = np.maximum.accumulate(capital)
        drawback_data = -((running_max - capital) / running_max)
        max_drawdown = np.min(drawback_data)

        return {
            'ir': ir,
            'sharpe': sharpe,
            'turnover': turnover,
            'returns': returns,
            'fitness': fitness,
            'margin': margin,
            'max_drawdown': max_drawdown
        }



# -*- coding: utf-8 -*-

import unittest

import numpy as np
import pandas as pd

from ultron.optimize.model.loader import load_model
from ultron.optimize.model.treemodel import RandomForestClassifier
from ultron.optimize.model.treemodel import RandomForestRegressor
from ultron.optimize.model.treemodel import ExtraTreesClassifier
from ultron.optimize.model.treemodel import ExtraTreesRegressor
from ultron.optimize.model.treemodel import BaggingClassifier
from ultron.optimize.model.treemodel import BaggingRegressor
from ultron.optimize.model.treemodel import AdaBoostClassifier
from ultron.optimize.model.treemodel import AdaBoostRegressor
from ultron.optimize.model.treemodel import GradientBoostingClassifier
from ultron.optimize.model.treemodel import GradientBoostingRegressor
from ultron.optimize.model.treemodel import DecisionTreeClassifier
from ultron.optimize.model.treemodel import DecisionTreeRegressor
from ultron.optimize.model.treemodel import StackingClassifier
from ultron.optimize.model.treemodel import StackingRegressor
from ultron.optimize.model.treemodel import VotingClassifier
from ultron.optimize.model.treemodel import VotingRegressor
from ultron.optimize.model.treemodel import XGBClassifier
from ultron.optimize.model.treemodel import XGBRegressor
from ultron.optimize.model.treemodel import XGBTrainer
from ultron.optimize.model.treemodel import LGBMClassifier
from ultron.optimize.model.treemodel import LGBMRegressor
from ultron.optimize.model.treemodel import LGBMTrainer
from ultron.optimize.model.linearmodel import LogisticRegression


class TreeModel(unittest.TestCase):

    def setUp(self):
        self.features = list('0123456789')
        self.x = pd.DataFrame(np.random.randn(1000, 10), columns=self.features)
        self.y = np.random.randn(1000)
        self.sample_x = pd.DataFrame(np.random.randn(100, 10),
                                     columns=self.features)
        self.sample_y = np.random.randn(100)

    def test_gradinet_boosting_regress_persistence(self):
        model = GradientBoostingRegressor(features=self.features)
        model.fit(self.x, self.y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_gradinet_boosting_classify_persistence(self):
        model = GradientBoostingClassifier(features=self.features)
        y = np.where(self.y > 0, 1, 0)
        model.fit(self.x, y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_ada_boost_regress_persistence(self):
        model = AdaBoostRegressor(features=self.features)
        model.fit(self.x, self.y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_ada_boost_classify_persistence(self):
        model = AdaBoostClassifier(features=self.features)
        y = np.where(self.y > 0, 1, 0)
        model.fit(self.x, y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_bagging_regress_persistence(self):
        model = BaggingRegressor(features=self.features)
        model.fit(self.x, self.y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_bagging_classify_persistence(self):
        model = BaggingClassifier(features=self.features)
        y = np.where(self.y > 0, 1, 0)
        model.fit(self.x, y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_extra_trees_regress_persistence(self):
        model = ExtraTreesRegressor(features=self.features)
        model.fit(self.x, self.y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_extra_trees_classify_persistence(self):
        model = ExtraTreesClassifier(features=self.features)
        y = np.where(self.y > 0, 1, 0)
        model.fit(self.x, y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_random_forest_regress_persistence(self):
        model = RandomForestRegressor(features=self.features)
        model.fit(self.x, self.y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_random_forest_classify_persistence(self):
        model = RandomForestClassifier(features=self.features)
        y = np.where(self.y > 0, 1, 0)
        model.fit(self.x, y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_decisiontree_regress_persistence(self):
        model = DecisionTreeRegressor(features=self.features)
        model.fit(self.x, self.y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_decisiontree_classify_persistence(self):
        model = DecisionTreeClassifier(features=self.features)
        y = np.where(self.y > 0, 1, 0)
        model.fit(self.x, y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    '''
    def test_stacking_regressor_persistence(self):
        m1 = RandomForestRegressor(features=self.features)
        m1.fit(self.x, self.y)
        m1_score = m1.score(self.sample_x, self.sample_y)

        m2 = ExtraTreesRegressor(features=self.features)
        m2.fit(self.x, self.y)
        m2_score = m2.score(self.sample_x, self.sample_y)

        m3 = BaggingRegressor(features=self.features)
        m3.fit(self.x, self.y)
        m3_score = m3.score(self.sample_x, self.sample_y)

        m4 = AdaBoostRegressor(features=self.features)
        m4.fit(self.x, self.y)
        m4_score = m4.score(self.sample_x, self.sample_y)


        m5 = GradientBoostingRegressor(features=self.features)
        m5.fit(self.x, self.y)
        m5_score = m1.score(self.sample_x, self.sample_y)

        model = StackingRegressor(features=self.features,
                regressors=[m1.device,  m2.device,
                                m3.device, m4.device, 
                                m5.device],
                    meta_regressor=LogisticRegression().device)
        model.fit(self.x, self.y)
        score = model.score(self.sample_x, self.sample_y)
        scores = np.array([m1_score, m2_score, m3_score, m4_score, m5_score])
        assert(score >= scores.min() - 0.3)

        desc = model.save()

        model1 = load_model(desc)
        score1 = model1.score(self.sample_x, self.sample_y)
        assert(score1 == score)
    '''

    def test_stacking_classify_persistence(self):
        y = np.where(self.y > 0, 1, 0)
        sample_y = np.where(self.sample_y > 0, 1, 0)
        m1 = RandomForestClassifier(features=self.features)
        m1.fit(self.x, y)
        m1_score = m1.score(self.sample_x, sample_y)

        m2 = ExtraTreesClassifier(features=self.features)
        m2.fit(self.x, y)
        m2_score = m2.score(self.sample_x, sample_y)

        m3 = BaggingClassifier(features=self.features)
        m3.fit(self.x, y)
        m3_score = m3.score(self.sample_x, sample_y)

        m4 = AdaBoostClassifier(features=self.features)
        m4.fit(self.x, y)
        m4_score = m4.score(self.sample_x, sample_y)

        m5 = GradientBoostingClassifier(features=self.features)
        m5.fit(self.x, y)
        m5_score = m1.score(self.sample_x, sample_y)

        model = StackingClassifier(features=self.features,
                                   classifiers=[
                                       m1.device, m2.device, m3.device,
                                       m4.device, m5.device
                                   ],
                                   meta_classifier=LogisticRegression().device)
        model.fit(self.x, y)
        score = model.score(self.sample_x, sample_y)
        scores = np.array([m1_score, m2_score, m3_score, m4_score, m5_score])
        assert (score >= scores.min() - 0.3)

        desc = model.save()

        model1 = load_model(desc)
        score1 = model1.score(self.sample_x, sample_y)
        assert (score1 == score)

    def test_voting_classify_persistence(self):
        y = np.where(self.y > 0, 1, 0)
        sample_y = np.where(self.sample_y > 0, 1, 0)
        m1 = RandomForestClassifier(features=self.features)
        m1.fit(self.x, y)
        m1_score = m1.score(self.sample_x, sample_y)

        m2 = ExtraTreesClassifier(features=self.features)
        m2.fit(self.x, y)
        m2_score = m2.score(self.sample_x, sample_y)

        m3 = BaggingClassifier(features=self.features)
        m3.fit(self.x, y)
        m3_score = m3.score(self.sample_x, sample_y)

        m4 = AdaBoostClassifier(features=self.features)
        m4.fit(self.x, y)
        m4_score = m4.score(self.sample_x, sample_y)

        m5 = GradientBoostingClassifier(features=self.features)
        m5.fit(self.x, y)
        m5_score = m1.score(self.sample_x, sample_y)

        model = VotingClassifier(features=self.features,
                                 estimators=[('m1', m1.device),
                                             ('m2', m2.device),
                                             ('m3', m3.device),
                                             ('m4', m4.device),
                                             ('m5', m5.device)],
                                 voting='soft')

        model.fit(self.x, y)
        score = model.score(self.sample_x, sample_y)
        scores = np.array([m1_score, m2_score, m3_score, m4_score, m5_score])
        assert (score >= scores.min() - 0.3)

        desc = model.save()

        model1 = load_model(desc)
        score1 = model1.score(self.sample_x, sample_y)
        assert (score1 == score)

    def test_xgb_regress_persistence(self):
        model = XGBRegressor(features=self.features)
        model.fit(self.x, self.y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_xgb_classify_persistence(self):
        model = XGBClassifier(features=self.features)
        y = np.where(self.y > 0, 1, 0)
        model.fit(self.x, y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_xgb_trainer_equal_classifier(self):
        model1 = XGBClassifier(n_estimators=100,
                               learning_rate=0.1,
                               max_depth=3,
                               features=self.features,
                               random_state=42)

        model2 = XGBTrainer(features=self.features,
                            objective='reg:logistic',
                            booster='gbtree',
                            tree_method='exact',
                            n_estimators=100,
                            learning_rate=0.1,
                            max_depth=3,
                            random_state=42)

        y = np.where(self.y > 0, 1, 0)
        model1.fit(self.x, y)
        model2.fit(self.x, y)

        predict1 = model1.predict(self.sample_x)
        predict2 = model2.predict(self.sample_x)
        predict2 = np.where(predict2 > 0.5, 1., 0.)
        np.testing.assert_array_almost_equal(predict1, predict2)

    def test_xgb_trainer_persistence(self):
        model = XGBTrainer(features=self.features,
                           objective='binary:logistic',
                           booster='gbtree',
                           tree_method='hist',
                           n_estimators=200)
        y = np.where(self.y > 0, 1, 0)
        model.fit(self.x, y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_lgbm_regress_persistence(self):
        model = LGBMRegressor(features=self.features)
        model.fit(self.x, self.y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    def test_lgbm_classify_persistence(self):
        model = LGBMClassifier(features=self.features)
        y = np.where(self.y > 0, 1, 0)
        model.fit(self.x, y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances,
                                             new_model.importances)

    '''
    def test_lgbm_trainer_persistence(self):
        model = LGBMTrainer(features=self.features,
                             objective='regression',
                             n_estimators=100,
                             learning_rate=0.1,
                             max_depth=3)
        y = np.where(self.y > 0, 1, 0)
        model.fit(self.x, y)

        desc = model.save()
        new_model = load_model(desc)
        self.assertEqual(model.features, new_model.features)

        np.testing.assert_array_almost_equal(model.predict(self.sample_x),
                                             new_model.predict(self.sample_x))
        np.testing.assert_array_almost_equal(model.importances, new_model.importances)
    '''


# -*- encoding:utf-8 -*-
"""
    买入择时策略因子基础模块
"""
import copy
from enum import Enum
from abc import ABCMeta, abstractmethod
from ultron.kdutils.lazy import LazyFunc
from ultron.ump.core.deprecated import Deprecated
from ultron.ump.core.fixes import six
from ultron.ump.core.base import ParamBase
from ultron.ump.slippage.buy_mean import SlippageBuyMean
from ultron.ump.trade.order import Order
from ultron.ump.ump.manager import Manager
from ultron.ump.trade.ml_feature import MlFeature
from ultron.ump.beta import position_base
from ultron.ump.beta.atr_position import AtrPosition


class BuyCallMixin(object):
    """
        混入类，混入代表买涨，不完全是期权中buy call的概念，
        只代表看涨正向操作，即期望买入后交易目标价格上涨，上涨带来收益
    """

    @LazyFunc
    def buy_type_str(self):
        """用来区别买入类型unique 值为call"""
        return "call"

    @LazyFunc
    def expect_direction(self):
        """期望收益方向，1.0即正向期望"""
        return 1.0


class BuyPutMixin(object):
    """
        混入类，混入代表买跌，应用场景在于期权，期货策略中，
        不完全是期权中buy put的概念，只代看跌反向操作，
        即期望买入后交易目标价格下跌，下跌带来收益
    """

    @LazyFunc
    def buy_type_str(self):
        """用来区别买入类型unique 值为put"""
        return "put"

    @LazyFunc
    def expect_direction(self):
        """期望收益方向，1.0即反向期望"""
        return -1.0


class FactorBuyBase(six.with_metaclass(ABCMeta, ParamBase)):
    """
        买入择时策略因子基类：每一个继承FactorBuyBase的子类必须混入一个方向类，
        且只能混入一个方向类，即具体买入因子必须明确买入方向，且只能有一个买入方向，
        一个因子不能同上又看涨又看跌，

        买入因子内部可容纳专属卖出因子和选股因子，即值针对本源生效的卖出因子和选股策略，
        且选股因子动态在择时周期内每月或者每周根据策略重新进行选股。
    """

    def __init__(self, capital, kl_pd, combine_kl_pd, benchmark, **kwargs):
        """
        :param capital:资金类AbuCapital实例化对象
        :param kl_pd:择时时段金融时间序列，pd.DataFrame对象
        :param combine_kl_pd:合并了之前一年时间序列的金融时间序列，pd.DataFrame对象
        :param benchmark: 交易基准对象，AbuBenchmark实例对象, 因子可有选择性使用，比如大盘对比等功能
        """
        # 择时金融时间序列走势数据
        self.kl_pd = kl_pd
        # 机器学习特征数据构建需要，详情见make_buy_order_ml_feature中构造特征使用
        self.combine_kl_pd = combine_kl_pd
        # 资金情况数据
        self.capital = capital
        # 交易基准对象，Benchmark实例对象, 因子可有选择性使用，比如大盘对比等功能
        self.benchmark = benchmark

        # 滑点类，默认SlippageBuyMean
        self._slippage_class_init(**kwargs)
        # 仓位管理，默认AbuAtrPosition
        self._position_class_init(**kwargs)

        # 构造ump对外的接口对象UmpManager
        self.ump_manger = Manager(factor=self)
        # 默认的factor_name，子类通过_init_self可覆盖更具体的名字
        self.factor_name = '{}'.format(self.__class__.__name__)

        # 忽略的交易日数量
        self.skip_days = 0
        # 是否封锁本源择时买入因子的执行，此值只通过本源择时买入因子附属的选股因子进行改变
        self.lock_factor = False
        # kwargs参数中其它设置赋予买入因子的参数
        self._other_kwargs_init(**kwargs)

        # 子类继续完成自有的构造
        self._init_self(**kwargs)

    def _position_class_init(self, **kwargs):
        """仓位管理类构建"""
        # 仓位管理，默认AtrPosition
        if position_base.g_default_pos_class is None:
            self.position_class = AtrPosition
            # 仓位管理类构建关键字参数默认空字典
            self.position_kwargs = dict()
        else:
            # 否则设置了全局仓位管理字典对象，弹出class
            default_pos_class = copy.deepcopy(
                position_base.g_default_pos_class)
            self.position_class = default_pos_class.pop('class')
            # 弹出class后剩下的就为其它关键字参数
            self.position_kwargs = default_pos_class

        if 'position' in kwargs:
            position = kwargs.pop('position', AtrPosition)
            if isinstance(position, six.class_types):
                # 如果position里面直接设置的是一个class，直接弹出
                self.position_class = position
            elif isinstance(position, dict):
                # 支持赋予字典结构 eg: {'class': AbuAtrPosition, 'atr_base_price': 20, 'atr_pos_base': 0.5}
                if 'class' not in position:
                    # 必须要有需要实例化的类信息
                    raise ValueError('position class key must name class !!!')
                position_cp = copy.deepcopy(position)
                # pop出类信息后剩下的都为类需要的参数
                self.position_class = position_cp.pop('class')
                # pop出class之后剩下的就class的构造关键字参数
                self.position_kwargs = position_cp
            else:
                raise TypeError(
                    '_position_class_init position type is {}'.format(
                        type(position)))

    def _slippage_class_init(self, **kwargs):
        """滑点类构建"""
        # 滑点类，默认AbuSlippageBuyMean
        self.slippage_class = kwargs.pop('slippage', SlippageBuyMean)
        # TODO 滑点类完善参数构建等需求

    def _other_kwargs_init(self, **kwargs):
        """
            kwargs参数中其它设置赋予买入因子的参数：
            可选参数win_rate：策略因子期望胜率（可根据历史回测结果计算得出）
            可选参数gains_mean：策略因子期望收益（可根据历史回测结果计算得出）
            可选参数gains_mean：策略因子期望亏损（可根据历史回测结果计算得出）

            可选参数stock_pickers：专属买入择时策略因子的选股因子序列，序列中对象为选股因子
            可选参数sell_factors： 专属买入择时策略因子的择时卖出因子序列，序列中对象为卖出因子
        """
        # 先处理过时的方法
        self._deprecated_kwargs_init(**kwargs)

        # 专属买入策略因子的选股周生效因子
        self.ps_week = []
        # 专属买入策略因子的选股月生效因子
        self.ps_month = []
        # 专属买入策略因子的选股因子
        stock_pickers = kwargs.pop('stock_pickers', [])
        for picker_class in stock_pickers:
            if picker_class is None:
                continue
            if 'class' not in picker_class:
                # 必须要有需要实例化的类信息
                raise ValueError('picker_class class key must name class !!!')
            picker_class_cp = copy.deepcopy(picker_class)
            # pop出类信息后剩下的都为类需要的参数
            class_fac = picker_class_cp.pop('class')
            # 专属买入策略因子独有可设置选股因子生效周期，默认一个月重新进行一次选股，设置week为一周
            pick_period = picker_class_cp.pop('pick_period', 'month')
            # 整合capital，benchmark等实例化因子对象
            picker = class_fac(self.capital, self.benchmark, **picker_class_cp)
            if pick_period == 'month':
                self.ps_month.append(picker)
            elif pick_period == 'week':
                self.ps_week.append(picker)
            else:
                raise ValueError('pick_period just support month|week!')

        # 专属买入策略因子的卖出因子策略，只针对本源择时买入因子生效
        self.sell_factors = []
        sell_factors = kwargs.pop('sell_factors', [])
        for factor_class in sell_factors:
            if factor_class is None:
                continue
            if 'class' not in factor_class:
                # 必须要有需要实例化的类信息
                raise ValueError('factor class key must name class !!!')
            factor_class_cp = copy.deepcopy(factor_class)
            # pop出类信息后剩下的都为类需要的参数
            class_fac = factor_class_cp.pop('class')
            # 整合capital，kl_pd等实例化因子对象
            factor = class_fac(self.capital, self.kl_pd, self.combine_kl_pd,
                               self.benchmark, **factor_class_cp)
            # 添加到本源卖出因子序列
            self.sell_factors.append(factor)

    def __str__(self):
        """打印对象显示：class name, slippage, position, kl_pd.info"""
        return '{}: slippage:{}, position:{} \nkl:\n{}'.format(
            self.__class__.__name__, self.slippage_class, self.position_class,
            self.kl_pd.info())

    __repr__ = __str__

    def make_buy_order(self, day_ind=-1):
        """
        根据交易发生的时间索引，依次进行交易订单生成，交易时间序列特征生成，
        决策交易是否拦截，生成特征学习数据，最终返回order，即订单生效
        :param day_ind: 交易发生的时间索引，即对应self.kl_pd.key
        """
        if day_ind == -1:
            # 默认模式下非高频，信号发出后，明天进行买入操作
            day_ind = self.today_ind

        order = Order()
        # Orde对象根据交易发生的时间索引生成交易订单

        order.fit_buy_order(day_ind, self)

        if order.order_deal:
            # 交易时间序列特征生成
            ml_feature_dict = self.make_buy_order_ml_feature(day_ind)
            # 决策交易是否被ump拦截还是可以放行
            block = self.make_ump_block_decision(ml_feature_dict)
            if block:
                return None

            # 如果交易即将成交，将生成的交易特征写入order的特征字段ml_features中，为之后使用机器学习计算学习特征，训练ump
            if order.ml_features is None:
                order.ml_features = ml_feature_dict
            else:
                order.ml_features.update(ml_feature_dict)
        # 返回order，订单生效
        return order

    def make_ump_block_decision(self, ml_feature_dict):
        """
        输入需要决策的当前买入交易特征通过ump模块的对外manager对交易进行决策，
        判断是否拦截买入交易，还是放行买入交易。子类可复写此方法，即子类策略因子实现
        自己的任意ump组合拦截方式，根据策略的拦截比例需要等等参数确定ump具体策略，
        且对于多种策略并行执行策略本身定制适合自己的拦截策略，提高灵活度
        :param ml_feature_dict: 需要决策的当前买入时刻交易特征dict
        :return: bool, 对ml_feature_dict所描述的交易特征是否进行拦截
        """
        return self.ump_manger.ump_block(ml_feature_dict)

    def make_buy_order_ml_feature(self, day_ind):
        """
        根据交易发生的时间索引构通过AbuMlFeature构建买入时刻的各个交易特征
        :param day_ind: 交易发生的时间索引，对应self.kl_pd.key
        :return:
        """
        return MlFeature().make_feature_dict(self.kl_pd,
                                             self.combine_kl_pd,
                                             day_ind,
                                             buy_feature=True)

    @abstractmethod
    def _init_self(self, **kwargs):
        """子类因子针对可扩展参数的初始化"""
        pass

    def read_fit_day(self, today):
        """
        在择时worker对象中做日交易的函数，亦可以理解为盘前的一些决策事件处理，
        内部会调用子类实现的fit_day函数
        :param today: 当前驱动的交易日金融时间序列数据
        :return: 生成的交易订单AbuOrder对象
        """
        if self.skip_days > 0:
            self.skip_days -= 1
            return None

        # 今天这个交易日在整个金融时间序列的序号
        self.today_ind = int(today.key)
        # 回测中默认忽略最后一个交易日
        if self.today_ind >= self.kl_pd.shape[0] - 1:
            return None

        return self.fit_day(today)

    def buy_tomorrow(self):
        """
        明天进行买入操作，比如突破策略使用了今天收盘的价格做为参数，发出了买入信号，
        需要进行明天买入操作，不能执行今天买入操作
        :return 生成的交易订单AbuOrder对象
        """
        return self.make_buy_order(self.today_ind)

    def buy_today(self):
        """
        今天即进行买入操作，需要不能使用今天的收盘数据等做为fit_day中信号判断，
        适合如非明确一天交易日时间或者特殊情况的买入信号
        :return 生成的交易订单Order对象
        """
        return self.make_buy_order(self.today_ind - 1)

    def _fit_pick_stock(self, today, pick_array):
        """买入因子专属选股因子执行，只要一个选股因子发出没有选中的信号，就封锁本源择时因子"""

        for picker in pick_array:
            pick_kl = self.past_today_kl(today, picker.xd)

            if pick_kl.empty or not picker.fit_pick(pick_kl, self.kl_pd.name):
                # 只要一个选股因子发出没有选中的信号，就封锁本源选股因子
                self.lock_factor = True
                return
            # 遍历所有专属选股因子后，都没有发出封锁因子信号，就打开因子
            self.lock_factor = False

    def fit_ps_week(self, today):
        """买入因子专属'周'选股因子执行，只要一个选股因子发出没有选中的信号，就封锁本源择时因子"""
        self._fit_pick_stock(today, self.ps_week)

    def fit_ps_month(self, today):
        """买入因子专属'月'选股因子执行，只要一个选股因子发出没有选中的信号，就封锁本源择时因子"""
        self._fit_pick_stock(today, self.ps_month)

    @abstractmethod
    def fit_day(self, today):
        """子类主要需要实现的函数，完成策略因子针对每一个交易日的买入交易策略"""
        pass

    def past_today_kl(self, today, past_day_cnt):
        """
            在fit_day, fit_month, fit_week等时间驱动经过的函数中通过传递今天的数据
            获取过去past_day_cnt天的交易日数据，返回为pd.DataFram数据
            :param today: 当前驱动的交易日金融时间序列数据
            :param past_day_cnt: int，获取今天之前过去past_day_cnt天的金融时间序列数据
        """
        end_ind = self.combine_kl_pd[self.combine_kl_pd.date ==
                                     today.date].key.values[0]
        start_ind = end_ind - past_day_cnt if end_ind - past_day_cnt > 0 else 0
        # 根据当前的交易日，切片过去一段时间金融时间序列
        return self.combine_kl_pd.iloc[start_ind:end_ind]

    def past_today_one_month(self, today):
        """套接past_today_kl，获取今天之前1个月交易日的金融时间序列数据"""
        # TODO 这里固定了值，最好使用env中的时间，如币类市场等特殊情况
        return self.past_today_kl(today, 20)

    def past_today_one_week(self, today):
        """套接past_today_kl，获取今天之前1周交易日的金融时间序列数据"""
        # TODO 这里固定了值，最好使用env中的时间，如币类市场等特殊情况
        return self.past_today_kl(today, 5)

    def past_today_one_year(self, today):
        """套接past_today_kl，获取今天之前1年交易日的金融时间序列数据"""
        # TODO 这里固定了值，最好使用env中的时间，如币类市场等特殊情况
        return self.past_today_kl(today, 250)

    def _deprecated_kwargs_init(self, **kwargs):
        """处理过时的初始化"""
        if 'win_rate' in kwargs and 'gains_mean' in kwargs and 'losses_mean' in kwargs:
            self._do_kelly_deprecated(**kwargs)

    @Deprecated(
        'kelly object now use dict to build, it will be remove next version!')
    def _do_kelly_deprecated(self, **kwargs):
        """针对kelly仓位管理过时方法的处理"""
        """
            因子可选择根据策略的历史回测设置胜率，期望收益，期望亏损，
            比如使用AbuKellyPosition，必须需要因子的胜率，期望收益，
            期望亏损参数，不要使用kwargs.pop('a', None)设置，因为
            暂时使用hasattr判断是否有设置属性
        """
        # 策略因子历史胜率win_rate, 策略因子历史期望收益gains_mean, 策略因子历史期望亏损losses_mean
        self.position_kwargs = {
            'win_rate': kwargs['win_rate'],
            'gains_mean': kwargs['gains_mean'],
            'losses_mean': kwargs['losses_mean']
        }

    """TODO: 使用check support方式查询是否支持fit_week，fit_month，上层不再使用hasattr去判断"""
    # def fit_week(self, *args, **kwargs):
    #     pass

    # def fit_month(self, *args, **kwargs):
    #     pass


class FactorBuyTD(FactorBuyBase):
    """很多策略中在fit_day中不仅仅使用今天的数据，经常使用昨天，前天数据，方便获取昨天，前天的封装"""

    def read_fit_day(self, today):
        """
        覆盖base函数完成:
        1. 为fit_day中截取昨天self.yesterday
        2. 为fit_day中截取前天self.bf_yesterday
        :param today: 当前驱动的交易日金融时间序列数据
        :return: 生成的交易订单AbuOrder对象
        """
        if self.skip_days > 0:
            self.skip_days -= 1
            return None

        # 今天这个交易日在整个金融时间序列的序号
        self.today_ind = int(today.key)
        # 回测中默认忽略最后一个交易日
        if self.today_ind >= self.kl_pd.shape[0] - 1:
            return None

        # 忽略不符合买入的天（统计周期内前2天，因为需要昨天和前天）
        if self.today_ind < 2:
            return None

        # 为fit_day中截取昨天
        self.yesterday = self.kl_pd.iloc[self.today_ind - 1]
        # 为fit_day中截取前天
        self.bf_yesterday = self.kl_pd.iloc[self.today_ind - 2]

        return self.fit_day(today)

    def _init_self(self, **kwargs):
        """raise NotImplementedError"""
        raise NotImplementedError('NotImplementedError _init_self')

    def fit_day(self, today):
        """raise NotImplementedError"""
        raise NotImplementedError('NotImplementedError fit_day')


class FactorBuyXD(FactorBuyBase):
    """以周期为重要参数的策略，xd代表参数'多少天'如已周期为参数可直接继承使用"""

    def read_fit_day(self, today):
        """
        覆盖base函数完成过滤统计周期内前xd天以及为fit_day中切片周期金融时间序列数据
        :param today: 当前驱动的交易日金融时间序列数据
        :return: 生成的交易订单AbuOrder对象
        """
        if self.skip_days > 0:
            self.skip_days -= 1
            return None

        # 今天这个交易日在整个金融时间序列的序号
        self.today_ind = int(today.key)
        # 回测中默认忽略最后一个交易日
        if self.today_ind >= self.kl_pd.shape[0] - 1:
            return None

        # 忽略不符合买入的天（统计周期内前xd天）
        if self.today_ind < self.xd - 1:
            return None

        # 完成为fit_day中切片周期金融时间序列数据
        self.xd_kl = self.kl_pd[self.today_ind - self.xd + 1:self.today_ind +
                                1]

        return self.fit_day(today)

    def buy_tomorrow(self):
        """
        覆盖base函数，明天进行买入操作，比如突破策略使用了今天收盘的价格做为参数，发出了买入信号，
        需要进行明天买入操作，不能执行今天买入操作，使用周期参数xd赋予skip_days
        :return 生成的交易订单AbuOrder对象
        """

        self.skip_days = self.xd
        return self.make_buy_order(self.today_ind)

    def buy_today(self):
        """
        覆盖base函数，今天即进行买入操作，需要不能使用今天的收盘数据等做为fit_day中信号判断，
        适合如比特币非明确一天交易日时间或者特殊情况的买入信号，，使用周期参数xd赋予skip_days
        :return 生成的交易订单AbuOrder对象
        """
        self.skip_days = self.xd
        return self.make_buy_order(self.today_ind - 1)

    def _init_self(self, **kwargs):
        """子类因子针对可扩展参数的初始化"""
        # 突破周期参数 xd， 比如20，30，40天...突破, 不要使用kwargs.pop('xd', 20), 明确需要参数xq
        self.xd = kwargs['xd']
        # 在输出生成的orders_pd中显示的名字
        self.factor_name = '{}:{}'.format(self.__class__.__name__, self.xd)

    def fit_day(self, today):
        """raise NotImplementedError"""
        raise NotImplementedError('NotImplementedError fit_day')



# -*- encoding:utf-8 -*-
"""
    symbol模块
"""
from ultron.ump.core.fixes import six
from ultron.kdutils.lazy import LazyFunc
from ultron.ump.core.env import EMarketTargetType, EMarketSubType


# noinspection PyProtectedMember
def code_to_symbol(code, rs=True):
    """
    解析code成Symbol,如果code中带有市场编码直接用该市场，否则进行查询所属市场信息，
    如果最后也没有发现symbol所在的市场，会向外raise ValueError
    :param code: str对象，代码 如：300104，sz300104，usTSLA
    :param rs: 没有匹配上是否对外抛异常，默认True
    :return: Symbol对象
    """
    from ultron.ump.market.futures import FuturesCn
    from ultron.ump.market.stock import SymbolCN, SymbolUS
    from ultron.ump.market.market import all_symbol

    if isinstance(code, Symbol):
        # code本身时symbol对象直接返回
        return code
    if not isinstance(code, six.string_types):
        # code必须是string_types
        raise TypeError('code must be string_types!!!，{} : type is {}'.format(
            code, type(code)))

    sub_market = None
    market = None
    # 尝试获取市场信息
    head = code[:2].lower()
    if head in [EMarketSubType.SH.value, EMarketSubType.SZ.value
                ] and code[2:].isdigit():
        # 市场信息匹配沪深股，查询子市场
        sub_market = EMarketSubType(head)
        market = EMarketTargetType.E_MARKET_TARGET_CN
    elif head == EMarketTargetType.E_MARKET_TARGET_CN.value and code[
            2:].isdigit():
        # 没指名A股子市场，使用如hs000001这种形式，需求获取symbol后查询子市场使用AbuSymbolCN().query_symbol_sub_market
        sub_market = EMarketSubType(SymbolCN().query_symbol_sub_market(
            code[2:]))
        market = EMarketTargetType.E_MARKET_TARGET_CN
    elif head == EMarketTargetType.E_MARKET_TARGET_US.value:
        # 市场信息匹配美股，查询子市场，AbuSymbolUS().query_symbol_sub_market
        sub_market = EMarketSubType(SymbolUS().query_symbol_sub_market(
            code[2:]))
        market = EMarketTargetType.E_MARKET_TARGET_US

    if market is not None and sub_market is not None:
        # 通过stock_code, market, sub_market构建Symbol对象
        stock_code = code[2:].upper()
        return Symbol(market, sub_market, stock_code)

    if code.isdigit():
        if len(code) == 6:
            # 6位全数字，匹配查询a股子市场
            market = EMarketTargetType.E_MARKET_TARGET_CN
            sub_market = EMarketSubType(
                SymbolCN().query_symbol_sub_market(code))
            return Symbol(market, sub_market, code)
        if rs:
            raise TypeError('cn symbol len = 6, hk symbol len = 5')
    elif code.isalpha() and code in all_symbol(
            EMarketTargetType.E_MARKET_TARGET_TC):
        # 全字母且匹配币类市场''btc', 'ltc'
        market = EMarketTargetType.E_MARKET_TARGET_TC
        sub_market = EMarketSubType.COIN
        return Symbol(market, sub_market, code)
    elif (code[0].isalpha()
          and code[1:].isdigit()) or (code[:2].isalpha()
                                      and code[2:].isdigit()):
        # 匹配国内期货市场symbol
        futures_code = code.upper()
        q_df = FuturesCn().query_symbol(futures_code)
        if q_df is not None:
            sub_market = EMarketSubType(q_df.exchange.values[0])
            market = EMarketTargetType.E_MARKET_TARGET_FUTURES_CN
            return Symbol(market, sub_market, futures_code)

    if rs:
        raise ValueError('arg code :{} format dt support'.format(code))


class Symbol(object):
    """统一所有市场的symbol，统一对外接口对象"""

    # 定义使用的sh大盘
    SH_INDEX = ['000001', '000300']
    # 定义使用的sz大盘
    SZ_INDEX = ['399001', '399006']

    def __init__(self, market, sub_market, symbol_code):
        """
        :param market: EMarketTargetType enum对象
        :param sub_market: EMarketSubType enum对象
        :param symbol_code: str对象，不包含市场信息的code
        """
        if isinstance(market, EMarketTargetType) and isinstance(
                sub_market, EMarketSubType):
            self.market = market
            self.sub_market = sub_market
            self.symbol_code = symbol_code
            self.source = None
        else:
            raise TypeError('market type error')

    def __str__(self):
        """打印对象显示：market， sub_market， symbol_code"""
        return '{}_{}:{}'.format(self.market.value, self.sub_market.value,
                                 self.symbol_code)

    __repr__ = __str__

    def __len__(self):
        """对象长度：拼接市场＋子市场＋code的字符串长度"""
        m_symbol = '{}_{}:{}'.format(self.market.value, self.sub_market.value,
                                     self.symbol_code)
        return len(m_symbol)

    @LazyFunc
    def value(self):
        """不同市场返回ABuSymbolPd.make_kl_df使用的symbol LazyFunc"""
        if self.market == EMarketTargetType.E_MARKET_TARGET_HK or self.market == EMarketTargetType.E_MARKET_TARGET_US:
            # hk,us eg: usTSLA, hk00836
            return '{}{}'.format(self.market.value, self.symbol_code)
        elif self.market == EMarketTargetType.E_MARKET_TARGET_CN:
            # cn eg: sh000001
            return '{}{}'.format(self.sub_market.value, self.symbol_code)
        # 其它市场直接返回symbol_code
        return self.symbol_code

    def is_a_stock(self):
        """判定是否a股symbol"""
        return self.market == EMarketTargetType.E_MARKET_TARGET_CN

    def is_sh_stock(self):
        """判定是否a股sh symbol"""
        return self.sub_market == EMarketSubType.SH

    def is_sz_stock(self):
        """判定是否a股sz symbol"""
        return self.sub_market == EMarketSubType.SZ

    def is_us_stock(self):
        """判定是否美股 symbol"""
        return self.market == EMarketTargetType.E_MARKET_TARGET_US

    def is_us_n_stock(self):
        """判定是否美股纽约交易所 symbol"""
        return self.sub_market == EMarketSubType.US_N

    def is_us_oq_stock(self):
        """判定是否美股纳斯达克交易所 symbol"""
        return self.sub_market == EMarketSubType.US_OQ

    def is_hk_stock(self):
        """判定是否港股 symbol"""
        return self.market == EMarketTargetType.E_MARKET_TARGET_HK

    def is_sh_index(self):
        """判定是否a股sh 大盘"""
        return self.is_sh_stock() and self.symbol_code in Symbol.SH_INDEX

    def is_sz_index(self):
        """判定是否a股sz 大盘"""
        return self.is_sz_stock() and self.symbol_code in Symbol.SZ_INDEX

    def is_a_index(self):
        """判定是否a股 大盘"""
        return self.is_sh_index() or self.is_sz_index()

    def is_index(self):
        """判定是否大盘"""
        return self.is_a_index()

    def is_futures(self):
        """判定是否期货symbol"""
        return self.market == EMarketTargetType.E_MARKET_TARGET_FUTURES_CN \
            or self.market == EMarketTargetType.E_MARKET_TARGET_FUTURES_GLOBAL

    def is_tc(self):
        """判定是否电子货币symbol"""
        return self.market == EMarketTargetType.E_MARKET_TARGET_TC



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd


class Universe(object):

    @classmethod
    def fix(cls, data):
        nrow, ncol = data.shape
        rval = np.full(shape=[nrow, ncol], fill_value=True)
        rval = pd.DataFrame(rval, index=data.index, columns=data.columns)

        products = [
            'A', 'AG', 'AL', 'AU', 'BU', 'C', 'CF', 'CS', 'CU', 'FG', 'HC',
            'I', 'J', 'JM', 'L', 'M', 'MA', 'NI', 'OI', 'P', 'PP', 'RB', 'RM',
            'RU', 'SR', 'TA', 'V', 'Y', 'ZC', 'ZN'
        ]

        for col in data.columns:
            if col not in products:
                rval[col] = False

        return rval

    @classmethod
    def single(cls, data, product):
        rval = pd.DataFrame(False, index=data.index, columns=data.columns)

        if isinstance(product, str):
            products = [product]
        else:
            products = product

        rval[products] = True

        return rval

    @classmethod
    def all(cls, data):
        assert isinstance(data, pd.DataFrame)
        rval = pd.DataFrame(True, index=data.index, columns=data.columns)
        return rval

    @classmethod
    def index_pair(cls, data):
        nrow, ncol = data.shape
        rval = np.full(shape=[nrow, ncol], fill_value=True)
        rval = pd.DataFrame(rval, index=data.index, columns=data.columns)
        products = ['IF', 'IM', 'IH', 'IC']
        for col in data.columns:
            if col not in products:
                rval[col] = False

        return rval


# -*- encoding:utf-8 -*-
"""示例ump多混特征模块"""

from ultron.ump.trade.ml_feature import FeatureDeg, FeaturePrice, FeatureJump, FeatureWave, FeatureAtr
from ultron.ump.ump.base import ump_main_make_xy, BuyUmpMixin
from ultron.ump.ump.main_base import MainBase
from ultron.ump.model.principles import Principles


class FullFiter(Principles):
    """
        内部类，只需要实现make_xy，且使用ump_main_make_xy装饰

        在__init__中通过：
                self.fiter_cls = self.get_fiter_class()
                self.fiter = self.fiter_cls(orders_pd=self.orders_pd, **kwarg)
        构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
    """

    @ump_main_make_xy
    def make_xy(self, **kwarg):
        """
        make_xy通过装饰器ump_main_make_xy进行二次包装
        这里只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df
        :param kwarg: ump_main_make_xy装饰器中使用kwarg
                      kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
                      kwargs['scaler']    做为可选关键字参数：控制在make_xy中返回的特征矩阵数据是否进行标准化处理

        :return: self.order_has_ret中使用filter选取特征列形成full_df
                 ump_main_make_xy装饰器在make_xy返回full_df后做转换matrix，形成x，y等工作
        """

        regex = 'result|{}|{}|{}|{}|{}'.format(
            '|'.join(FeatureDeg().get_feature_ump_keys(ump_cls=MainFull)),
            '|'.join(FeaturePrice().get_feature_ump_keys(ump_cls=MainFull)),
            '|'.join(FeatureWave().get_feature_ump_keys(ump_cls=MainFull)),
            '|'.join(FeatureAtr().get_feature_ump_keys(ump_cls=MainFull)),
            '|'.join(FeatureJump().get_feature_ump_keys(ump_cls=MainFull)))
        """
            regex = 'result|buy_deg_ang42|buy_deg_ang252|buy_deg_ang60|buy_deg_ang21|buy_price_rank120|
            buy_price_rank90|buy_price_rank60|buy_price_rank252|buy_wave_score1|buy_wave_score2|buy_wave_score3
            |buy_atr_std|buy_diff_down_days|buy_jump_up_power|buy_diff_up_days|buy_jump_down_power'
        """

        # noinspection PyUnresolvedReferences
        full_df = self.order_has_ret.filter(regex=regex)
        """
            eg: full_df
                        result  buy_deg_ang42  buy_deg_ang252  buy_deg_ang60  \
            2014-09-24       0          3.378           3.458          3.458
            2014-10-24       0          0.191           2.889          2.809
            2014-10-29       1         -2.026          16.689         -0.761
            2014-10-29       1         -3.427         -11.956         -8.296
            2014-10-29       1         -2.915          39.469         -6.043
            2014-10-29       1         -2.026          16.689         -0.761
            ...            ...            ...             ...            ...
            2016-04-14       0          4.611          18.428          3.134
            2016-04-15       0          4.238         -13.247          4.693
            2016-04-15       0          4.238         -13.247          4.693
            2016-04-29       1          1.445          16.266          4.615
            2016-04-29       1          1.445          16.266          4.615

                        buy_deg_ang21  buy_price_rank120  buy_price_rank90  \
            2014-09-24          1.818              1.000             1.000
            2014-10-24         -1.089              1.000             1.000
            2014-10-29          1.980              1.000             1.000
            2014-10-29          6.507              0.925             0.900
            2014-10-29          7.046              0.962             0.950
            2014-10-29          1.980              1.000             1.000
            ...                   ...                ...               ...
            2016-04-14          0.733              0.875             0.878
            2016-04-15          1.162              0.775             0.733
            2016-04-15          1.162              0.775             0.733
            2016-04-29         -1.115              1.000             1.000
            2016-04-29         -1.115              1.000             1.000

                        buy_price_rank60  buy_price_rank252  buy_wave_score1  \
            2014-09-24             1.000              1.000            0.287
            2014-10-24             1.000              1.000            0.596
            2014-10-29             1.000              1.000            0.444
            2014-10-29             0.883              0.750           -0.173
            2014-10-29             0.925              0.982            0.031
            2014-10-29             1.000              1.000            0.444
            ...                      ...                ...              ...
            2016-04-14             0.967              0.940           -0.030
            2016-04-15             1.000              0.631            0.596
            2016-04-15             1.000              0.631            0.596
            2016-04-29             1.000              1.000            0.743
            2016-04-29             1.000              1.000            0.743

                        buy_wave_score2  buy_wave_score3  buy_atr_std  \
            2014-09-24            0.234            0.218        0.226
            2014-10-24            0.488            0.449        0.146
            2014-10-29            0.338            0.329        0.538
            2014-10-29           -0.202           -0.203        0.558
            2014-10-29           -0.128           -0.173        0.063
            2014-10-29            0.338            0.329        0.538
            ...                     ...              ...          ...
            2016-04-14            0.081            0.170        0.049
            2016-04-15            0.753            0.800        0.135
            2016-04-15            0.753            0.800        0.135
            2016-04-29            0.840            0.918        0.838
            2016-04-29            0.840            0.918        0.838

                        buy_jump_down_power  buy_diff_down_days  buy_jump_up_power  \
            2014-09-24                0.000                   0              3.344
            2014-10-24                0.000                   0              3.344
            2014-10-29               -1.109                 278              2.920
            2014-10-29                0.000                   0              1.283
            2014-10-29               -1.522                  75              3.727
            2014-10-29               -1.109                 278              2.920
            ...                         ...                 ...                ...
            2016-04-14               -1.764                 100              1.158
            2016-04-15               -1.455                 101              1.075
            2016-04-15               -1.455                 101              1.075
            2016-04-29               -1.178                  24              3.259
            2016-04-29               -1.178                  24              3.259

                        buy_diff_up_days
            2014-09-24                61
            2014-10-24                91
            2014-10-29                95
            2014-10-29               243
            2014-10-29               238
            2014-10-29                95
            ...                      ...
            2016-04-14                71
            2016-04-15                58
            2016-04-15                58
            2016-04-29                 0
            2016-04-29                 0
        """
        return full_df


class MainFull(MainBase, BuyUmpMixin):
    """主裁多混特征类，UltronUmpMainBase子类，混入BuyUmpMixin，做为买入ump类"""

    def get_predict_col(self):
        """
        主裁多混特征keys：
            ['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21', 'buy_price_rank120',
            'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252',
            'buy_wave_score1', 'buy_wave_score2', 'buy_wave_score3', 'buy_atr_std',
            'buy_diff_down_days', 'buy_jump_up_power', 'buy_diff_up_days', 'buy_jump_down_power']

        :return: ['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21', 'buy_price_rank120',
                'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252',
                'buy_wave_score1', 'buy_wave_score2', 'buy_wave_score3', 'buy_atr_std',
                'buy_diff_down_days', 'buy_jump_up_power', 'buy_diff_up_days', 'buy_jump_down_power']
        """
        cols = FeatureDeg().get_feature_ump_keys(ump_cls=MainFull)
        cols.extend(FeaturePrice().get_feature_ump_keys(ump_cls=MainFull))
        cols.extend(FeatureWave().get_feature_ump_keys(ump_cls=MainFull))
        cols.extend(FeatureAtr().get_feature_ump_keys(ump_cls=MainFull))
        cols.extend(FeatureJump().get_feature_ump_keys(ump_cls=MainFull))
        return cols

    def get_fiter_class(self):
        """
        主裁多混特征返回的UltronMLPd子类：UltronUmpMainFull.UmpFullFiter
        :return: UltronUmpMainFull.UmpFullFiter
        """
        return FullFiter

    @classmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法：return 'full_main'
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见UltronUmpManager中extend_ump_block方法
        """
        return 'full_main'



# -*- coding: utf-8 -*-
import os, pdb, itertools, copy, datetime, sys, json

os.environ['ULTRON_DATA'] = 'keim'

import numpy as np
import pandas as pd
from ultron.env import *
from ultron.utilities.utils import NumpyEncoder
from ultron.optimize.geneticist.engine import Engine
from ultron.optimize.model.treemodel import XGBRegressor
from ultron.optimize.model.treemodel import LGBMRegressor
from ultron.optimize.model.treemodel import RandomForestRegressor

enable_example_env()

train_data = pd.read_csv(os.path.join(g_project_data, 'train_datas.csv'),
                         index_col=0)


class Parameter(object):

    @classmethod
    def RandomForestRegressor(cls, **kwargs):
        return {
            'n_estimators': [i for i in range(80, 180, 10)],
            'max_depth': [i for i in range(3, 5, 2)],
            'max_features': ['auto', 'sqrt', 'log2']
        }

    @classmethod
    def ExtraTreesRegressor(cls, **kwargs):
        return {
            'n_estimators': [i for i in range(80, 180, 10)],
            'max_depth': [i for i in range(3, 5, 2)],
            'max_features': ['auto', 'sqrt', 'log2']
        }

    @classmethod
    def BaggingRegressor(cls, **kwargs):
        return {
            'n_estimators': [i for i in range(80, 180, 10)],
            'max_depth': [i for i in range(3, 5, 2)],
            'max_features': ['auto', 'sqrt', 'log2']
        }

    @classmethod
    def XGBRegressor(cls, **kwargs):
        return {
            'max_depth': [i for i in range(6, 15, 2)],
            'n_estimators': [i for i in range(50, 150, 10)],
            'learning_rate': [(i / 100) for i in range(1, 10, 2)]
        }

    @classmethod
    def LGBMRegressor(cls, **kwargs):
        return {
            'max_depth': [i for i in range(6, 15, 2)],
            'n_estimators': [i for i in range(50, 150, 10)],
            'learning_rate': [(i / 100) for i in range(1, 10, 2)]
        }


params_sets = {
    'XGBRegressor': Parameter.XGBRegressor(),
    'RandomForestRegressor': Parameter.RandomForestRegressor(),
    'LGBMRegressor': Parameter.LGBMRegressor(),
}

model_sets = ['XGBRegressor', 'RandomForestRegressor', 'LGBMRegressor']

features = [
    col for col in train_data.columns if col not in
    ['trade_date', 'code', 'value', 'signal', 'inventory', 'profitratio']
]

X = train_data[['trade_date', 'code'] + features].set_index(
    ['trade_date', 'code']).fillna(0)
Y = train_data[['trade_date', 'code',
                'signal']].set_index(['trade_date', 'code']).fillna(0)
Y = pd.DataFrame(np.random.randn(len(X)), index=Y.index, columns=['signal'])


def save_model(gen, rootid, sessionid, run_details):  ## 每一代优秀模型回调
    res = []
    pdb.set_trace()
    for detail in run_details:
        features = detail._identification
        model_name = detail._model_name
        params = json.dumps(detail._params, cls=NumpyEncoder)
        fitness = detail._raw_fitness
        method = detail._init_method
        res.append({
            'features': features,
            'model_name': model_name,
            'params': params,
            'fitness': fitness,
            'gen': gen,
            'method': method
        })


custom_params = {
    'begin_date': '2020-01-01',
    'end_date': '2022-01-01',
    'yields': 'ret',
    'rootid': '10001',
    'thresh': 0.98
}

gentic = Engine(model_sets=model_sets,
                params_sets=params_sets,
                convergence=0.002,
                population_size=20,
                tournament_size=5,
                p_point_mutation=0.7,
                p_crossover=0.2,
                standard_score=0.2,
                stopping_criteria=-100,
                custom_params=custom_params,
                rootid='100001',
                greater_is_better=False,
                save_model=save_model)

gentic.train(features, X=X, Y=Y, mode='mse', n_splits=2)  # mode 为模型评估方式



# -*- coding: utf-8 -*-

import unittest

import numpy as np

from ultron.optimize.meanvariancebuilder import mean_variance_builder
from ultron.optimize.meanvariancebuilder import target_vol_builder


class MeanVarianceBuild(unittest.TestCase):

    def test_mean_variance_builder(self):
        er = np.array([0.01, 0.02, 0.03])
        cov = np.array([[0.02, 0.01, 0.02],
                        [0.01, 0.02, 0.03],
                        [0.02, 0.03, 0.02]])
        ids_var = np.diag([0.01, 0.02, 0.03])
        cov += ids_var

        bm = np.array([0.3, 0.3, 0.4])
        lbound = np.array([0., 0., 0.])
        ubound = np.array([0.4, 0.4, 0.5])

        risk_exposure = np.array([[1., 1., 1.],
                                  [1., 0., 1.]]).T
        risk_target = (np.array([bm.sum(), 0.3]), np.array([bm.sum(), 0.7]))

        model = dict(cov=cov, factor_cov=None, factor_loading=None, idsync=None)
        status, _, x = mean_variance_builder(er, model, bm, lbound, ubound, risk_exposure,
                                             risk_target)

        self.assertTrue(status == 'optimal')
        self.assertAlmostEqual(x.sum(), bm.sum())
        self.assertTrue(np.all(x <= ubound + 1.e-6))
        self.assertTrue(np.all(x >= lbound) - 1.e-6)
        self.assertTrue(np.all(x @ risk_exposure <= risk_target[1] + 1.e-6))
        self.assertTrue(np.all(x @ risk_exposure >= risk_target[0] - 1.e-6))
        np.testing.assert_array_almost_equal(x, [0.1, 0.4, 0.5])

    def test_mean_variance_builder_without_constraints(self):
        er = np.array([0.01, 0.02, 0.03])
        cov = np.array([[0.02, 0.01, 0.02],
                        [0.01, 0.02, 0.03],
                        [0.02, 0.03, 0.02]])
        ids_var = np.diag([0.01, 0.02, 0.03])
        cov += ids_var

        bm = np.array([0., 0., 0.])

        model = dict(cov=cov, factor_cov=None, factor_loading=None, idsync=None)
        status, _, x = mean_variance_builder(er, model, bm, None, None, None, None, lam=1)
        np.testing.assert_array_almost_equal(x, np.linalg.inv(cov) @ er)
    
    def test_mean_variance_builder_with_none_unity_lambda(self):
        er = np.array([0.01, 0.02, 0.03])
        cov = np.array([[0.02, 0.01, 0.02],
                        [0.01, 0.02, 0.03],
                        [0.02, 0.03, 0.02]])
        ids_var = np.diag([0.01, 0.02, 0.03])
        cov += ids_var

        bm = np.array([0.3, 0.3, 0.4])
        lbound = np.array([0., 0., 0.])
        ubound = np.array([0.4, 0.4, 0.5])

        risk_exposure = np.array([[1., 1., 1.],
                                  [1., 0., 1.]]).T
        risk_target = (np.array([bm.sum(), 0.3]), np.array([bm.sum(), 0.7]))

        model = dict(cov=cov, factor_cov=None, factor_loading=None, idsync=None)
        status, _, x = mean_variance_builder(er, model, bm, lbound, ubound, risk_exposure,
                                             risk_target, lam=100)

        self.assertTrue(status == 'optimal')
        self.assertAlmostEqual(x.sum(), bm.sum())
        self.assertTrue(np.all(x <= ubound + 1.e-6))
        self.assertTrue(np.all(x >= lbound) - 1.e-6)
        self.assertTrue(np.all(x @ risk_exposure <= risk_target[1] + 1.e-6))
        self.assertTrue(np.all(x @ risk_exposure >= risk_target[0] - 1.e-6))
        np.testing.assert_array_almost_equal(x, [0.2950, 0.3000, 0.4050])

    def test_target_vol_builder(self):
        er = np.array([0.1, 0.2, 0.3])
        cov = np.array([[0.05, 0.01, 0.02],
                        [0.01, 0.06, 0.03],
                        [0.02, 0.03, 0.07]])

        lbound = np.array([0., 0., 0.])
        ubound = np.array([0.8, 0.8, 0.8])

        bm = np.array([0.3, 0.3, 0.3])

        risk_exposure = np.array([[1., 1., 1.]]).T
        risk_target = (np.array([bm.sum()]), np.array([bm.sum()]))

        model = dict(cov=cov, factor_cov=None, factor_loading=None, idsync=None)
        status, _, x = target_vol_builder(er, model, bm, lbound, ubound, risk_exposure, risk_target,
                                          0.1)
        self.assertTrue(status == 'optimal')
        self.assertTrue(np.all(x <= ubound + 1.e-6))
        self.assertTrue(np.all(x >= lbound) - 1.e-6)
        self.assertTrue(np.all(x @ risk_exposure <= risk_target[1] + 1.e-6))
        self.assertTrue(np.all(x @ risk_exposure >= risk_target[0] - 1.e-6))
        np.testing.assert_array_almost_equal(x, [-0.3, -0.10919033, 0.40919033] + bm)

    
        


        




# -*- coding: utf-8 -*-
import json
import datetime
import gevent
import importlib
import hashlib
from gevent.queue import Queue
from ultron.utilities.redis.redis_client import RedisClient
from ultron.cluster.work.extern_modules.modules_info import moddules_info
from ultron.utilities.short_uuid import unique_machine
from ultron.config import config_setting


class WorkEngine(object):

    def __init__(self, **kwargs):
        self._module_dict = {}
        self._redis_client = RedisClient(host=config_setting.queue_host,
                                         port=config_setting.queue_port,
                                         password=config_setting.queue_pwd)
        self._secret_key = 'd6f89b09'
        self._wid = unique_machine
        self._queue_list = ['ultron:work:work_id:' + str(self._wid)]
        #生成token
        self._token = hashlib.sha1(
            (self._secret_key +
             self._wid.replace('-', '')).encode()).hexdigest()
        self._task_queue = Queue()  #
        self.init_modules()
        gevent.spawn(self._get_task)
        gevent.spawn(self._dispatch_task)
        gevent.spawn(self._heart_tick)
        gevent.sleep()

    def init_modules(self):
        l = json.loads(moddules_info)
        for setting in l:
            self.load_modules(setting)

    def load_modules(self, setting):
        name = setting['name']
        is_effective = setting['isEffective']
        if is_effective == 0:
            return
        module_name = 'ultron.cluster.work.extern_modules.' + name + '.module'
        try:
            module = importlib.import_module(module_name)
            if 'Module' in dir(module):
                strategy_class = module.__getattribute__('Module')
                self._module_dict[name] = strategy_class(
                    name, self._wid, self._token, self._redis_client)
                print('module %s loading' % (name))
                if name == 'login':
                    self._module_dict[name].login_master()
        except Exception as e:
            print('Failed to import module:%s:[%s]' % (name, str(e)))

    def _heart_tick(self):
        last_time = datetime.datetime.now()
        while True:
            now_time = datetime.datetime.now()
            if (now_time - last_time).seconds > 20:  #发送心跳包
                task = {'name': 'login', 'opcode': 'heart_tick'}
                self._module_dict[task['name']].process_respone(task)
                last_time = now_time
            gevent.sleep(.3)

    def _get_task(self):
        while True:
            for queue in self._queue_list:
                task_all = self._redis_client.hmgetall(queue)
                task_list = task_all[0]
                self._redis_client.hmdel(queue, task_list.keys())
                for tid, task in task_list.items():
                    self._task_queue.put(json.loads(task))
            gevent.sleep(.3)

    #用于处理各个节点登录
    def _dispatch_task(self):
        while True:
            while not self._task_queue.empty():
                task = self._task_queue.get()
                space_name = str(task.get('name'))
                if space_name in self._module_dict:
                    self._module_dict[space_name].process_respone(task)
            gevent.sleep(.3)



# -*- coding: utf-8 -*-

from ultron.sentry.Math.Distributions.NormalDistribution import NormalDistribution
from ultron.sentry.Math.Distributions.NormalDistribution import CumulativeNormalDistribution
from ultron.sentry.Math.Distributions.NormalDistribution import InverseCumulativeNormal

__all__ = ['NormalDistribution',
           'CumulativeNormalDistribution',
           'InverseCumulativeNormal']



import numpy as np
import pandas as pd
from ultron.factor.data.processing import factor_processing
from ultron.factor.data.standardize import standardize
from ultron.factor.data.winsorize import winsorize_normal
from ultron.factor.data.neutralize import neutralize
from ultron.factor.combine.kutil import calc_ic


class IC_Weighted(object):

    def __init__(self, result_type):
        self._result_type = result_type
        self._industry_styles = [
            'Bank', 'RealEstate', 'Health', 'Transportation', 'Mining',
            'NonFerMetal', 'HouseApp', 'LeiService', 'MachiEquip', 'BuildDeco',
            'CommeTrade', 'CONMAT', 'Auto', 'Textile', 'FoodBever',
            'Electronics', 'Computer', 'LightIndus', 'Utilities', 'Telecom',
            'AgriForest', 'CHEM', 'Media', 'IronSteel', 'NonBankFinan',
            'ELECEQP', 'AERODEF', 'Conglomerates'
        ]
        self._risk_styles = [
            'BETA', 'MOMENTUM', 'SIZE', 'EARNYILD', 'RESVOL', 'GROWTH', 'BTOP',
            'LEVERAGE', 'LIQUIDTY', 'SIZENL'
        ]

    #暂时写此处，去极值
    def winsorize(self,
                  total_data,
                  method='sigma',
                  limits=(3.0, 3.0),
                  drop=True):
        se = total_data.copy()
        if method == 'quantile':
            down, up = se.quantile([limits[0], 1.0 - limits[1]])
        elif method == 'sigma':
            std, mean = se.std(), se.mean()
            down, up = mean - limits[0] * std, mean + limits[1] * std

        if drop:
            se[se < down] = np.NaN
            se[se > up] = np.NaN
        else:
            se[se < down] = down
            se[se > up] = up
        return se

    #中性化
    def neutralize(self, se, risk_df):
        se_index = se.index
        se = se.dropna()
        risk = risk_df.loc[se.index]

        # use numpy for neu, which is faster
        x = np.linalg.lstsq(risk.values, np.matrix(se).T, rcond=None)[0]
        se_neu = se - risk.dot(x)[0]

        # # use statsmodels for neu
        # x = risk.values
        # y = se.values
        # model = sm.OLS(y, x)
        # results = model.fit()
        # se_neu = pd.Series(data=y - x.dot(results.params), index=se.index)

        return se_neu.reindex(se_index)

    #标准化
    def standardize(self, total_data):
        try:
            res = (total_data - np.nanmean(total_data)) / np.nanstd(total_data)
        except:
            res = pd.Series(data=np.NaN, index=total_data.index)
        return res

    def ic_calc(self,
                factor_data,
                forward_returns,
                risk_data,
                factor_name,
                risk_styles=None,
                method='quantile',
                up_limit=0.025,
                down_limit=0.025,
                return_col_name='nxt1_ret',
                ic_type='spearman'):
        fac_list = [factor_name]
        factors_se = factor_data.merge(risk_data.reset_index(),
                                       on=['code', 'trade_date'])
        ndiff_field = [
            i for i in list(set(factors_se.columns)) if i not in fac_list
        ]
        risk_styles = self._risk_styles if risk_styles is None else risk_styles
        alpha_res = []
        grouped = factors_se.groupby(['trade_date'])
        for k, g in grouped:
            ret_preprocess = neutralize(
                g[self._industry_styles + risk_styles].values.astype(float),
                factor_processing(g[fac_list].fillna(0).values,
                                  pre_process=[winsorize_normal, standardize]))
            f = pd.DataFrame(ret_preprocess, columns=fac_list)
            for k in ndiff_field:
                f[k] = g[k].values
            alpha_res.append(f)
        factor_se = pd.concat(alpha_res).set_index(['trade_date',
                                                    'code'])[fac_list]
        forward_returns = forward_returns.set_index(['trade_date', 'code'])
        return calc_ic(factor_df=factor_se.reset_index(),
                       return_df=forward_returns,
                       factor_list=fac_list,
                       return_col_name=return_col_name,
                       ic_type=ic_type)

    def run(self,
            factor_data,
            forward_returns,
            risk_data,
            factor_name,
            is_neutralize=True,
            risk_styles=None,
            method='quantile',
            up_limit=0.025,
            down_limit=0.025,
            return_col_name='nxt1_ret',
            ic_type='spearman'):
        ic_serliaze = self.ic_calc(factor_data=factor_data,
                                   forward_returns=forward_returns,
                                   risk_data=risk_data,
                                   factor_name=factor_name,
                                   risk_styles=risk_styles,
                                   method=method,
                                   up_limit=up_limit,
                                   down_limit=down_limit,
                                   return_col_name=return_col_name,
                                   ic_type=ic_type)[factor_name].dropna()
        return ic_serliaze.values.mean(
        ) if self._result_type == 'ic' else ic_serliaze.values.mean(
        ) / ic_serliaze.values.std()



# -*- coding: utf-8 -*-
import unittest

import numpy as np
import pandas as pd

from ultron.optimize.longshortbulder import long_short_builder

class LongShortBuild(unittest.TestCase):
    def setUp(self):
        self.x = np.random.randn(3000, 10)
        self.groups = np.random.randint(10, 40, size=3000)
        choices = np.random.choice(3000, 100, replace=False)
        self.masks = np.full(3000, False, dtype=bool)
        self.masks[choices] = True

    def test_long_short_build(self):
        x = self.x[:,0].flatten()
        calc_weights = long_short_builder(x).flatten()
        expected_weights = x / np.abs(x).sum()
        np.testing.assert_array_almost_equal(calc_weights, expected_weights)

        calc_weights = long_short_builder(self.x, leverage=2)
        expected_weights = self.x / np.abs(self.x).sum(axis=0) * 2
        np.testing.assert_array_almost_equal(calc_weights, expected_weights)

    def test_long_short_build_with_group(self):
        x = self.x[:, 0].flatten()
        calc_weights = long_short_builder(x, groups=self.groups).flatten()
        expected_weights = pd.Series(x).groupby(self.groups).apply(lambda s: s / np.abs(s).sum())
        np.testing.assert_array_almost_equal(calc_weights, expected_weights)

        calc_weights = long_short_builder(self.x, groups=self.groups)
        expected_weights = pd.DataFrame(self.x).groupby(self.groups).apply(
            lambda s: s / np.abs(s).sum(axis=0))
        np.testing.assert_array_almost_equal(calc_weights, expected_weights)

    def test_long_short_build_with_masks(self):
        x = self.x[:, 0].flatten()
        calc_weights = long_short_builder(x, masks=self.masks, leverage=1.).flatten()
        self.assertAlmostEqual(calc_weights.sum(), 0.)

        masked_x = x.copy()
        masked_x[self.masks] = 0.
        masked_x[~self.masks] = masked_x[~self.masks] - masked_x[~self.masks].mean()
        expected_weights = masked_x / np.abs(masked_x).sum()
        np.testing.assert_array_almost_equal(calc_weights, expected_weights)










import pickle, pdb
import os, copy, sys, datetime

os.environ['ULTRON_DATA'] = 'keim'
os.environ['IGNORE_WARNINGS'] = '0'
import numpy as np
import pandas as pd

sys.path.insert(0, os.path.abspath('../'))

from ultron.env import *
from ultron.ump.core import env
from ultron.ump.indicator.atr import atr14, atr21
from ultron.ump.market.symbol_pd import _benchmark
from ultron.ump.factor import FactorPreAtrNStop, FactorAtrNStop, FactorCloseAtrNStop, \
    FactorBuyBreak, FactorBuyPutBreak
from ultron.ump.metrics import grid_helper as GridHelper
from ultron.ump.metrics.grid_search import GridSearch
from ultron.ump.metrics.metrics_base import MetricsBase
from ultron.ump.metrics.score import WrsmScorer, BaseScorer
from ultron.ump.metrics.base import MetricsDemo
from matplotlib import pyplot as plt
import seaborn as sns

score_tuple_array = pickle.load(open('score_tuple_array.kpl', 'rb'))


class DemoScorer(BaseScorer):

    def _init_self_begin(self, *arg, **kwargs):
        """胜率，策略收益，alpha组成select_score_func"""
        pdb.set_trace()
        self.select_score_func = lambda metrics: [
            metrics.win_rate, metrics.algorithm_period_returns, metrics.alpha
        ]
        self.columns_name = ['win_rate', 'returns', 'alpha']
        self.weights_cnt = len(self.columns_name)

    def _init_self_end(self, *arg, **kwargs):
        """
        _init_self_end这里一般的任务是将score_pd中需要反转的反转，默认是数据越大越好，有些是越小越好，
        类似make_scorer(xxx, greater_is_better=True)中的参数greater_is_better的作用：

                            sign = 1 if greater_is_better else -1
        """
        self.score_pd['alpha'] = -self.score_pd['alpha']  ## 此处为假设条件


scorer = DemoScorer(score_tuple_array, metrics_class=MetricsDemo)
# 返回按照评分排序后的队列
scorer_returns_max = scorer.fit_score()
scorer.score_pd.sort_values(by='alpha').tail()



# -*- encoding:utf-8 -*-
"""示例ump边裁特征单混模块"""

from ultron.ump.trade.ml_feature import FeatureDeg, FeaturePrice, FeatureJump, FeatureWave, FeatureAtr
from ultron.ump.ump.base import ump_edge_make_xy, BuyUmpMixin
from ultron.ump.ump.edge_base import EdgeBase
from ultron.ump.model.principles import Principles


class MulFiter(Principles):
    """
        内部类，只需要实现make_xy，且使用ump_edge_make_xy装饰

        在边裁__init__中通过：
                self.fiter_cls = self.get_fiter_class()
                self.fiter = self.fiter_cls(orders_pd=orders_pd, **kwarg)
        构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
    """

    @ump_edge_make_xy
    def make_xy(self, **kwarg):
        """
         make_xy通过装饰器ump_edge_make_xy进行包装，调用make_xy前将有交易结果的单子进行筛选：
         order_has_ret = orders_pd[orders_pd['result'] != 0]，且赋予self.order_has_ret
         make_xy只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df

        :param kwarg: ump_edge_make_xy装饰器中使用kwarg
                      kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
        :return: self.order_has_ret中使用filter选取特征列形成mul_df
                 ump_edge_make_xy装饰器在make_xy返回mul_df后做转换matrix，形成x，y等工作
        """

        filter_list = ['profit', 'profit_cg']
        # ['profit', 'profit_cg', 'buy_deg_ang21', 'buy_price_rank252', 'buy_wave_score3', 'buy_atr_std']
        filter_list.extend([
            FeatureDeg().get_feature_ump_keys(ump_cls=EdgeMul)[-1],
            FeaturePrice().get_feature_ump_keys(ump_cls=EdgeMul)[-1],
            FeatureWave().get_feature_ump_keys(ump_cls=EdgeMul)[-1],
            FeatureAtr().get_feature_ump_keys(ump_cls=EdgeMul)[-1]
        ])
        # noinspection PyUnresolvedReferences
        mul_df = self.order_has_ret.filter(filter_list)
        """
            eg: mul_df
                           profit  profit_cg  buy_deg_ang21  buy_price_rank252  \
            2014-09-24  -22618.04    -0.0566          1.818              1.000
            2014-10-24  -29690.28    -0.0742         -1.089              1.000
            2014-10-29   18959.19     0.0542          1.980              1.000
            2014-10-29  148209.36     0.5022          6.507              0.750
            2014-10-29   24867.60     0.0952          7.046              0.982
            2014-10-29   18959.19     0.0542          1.980              1.000
            2014-11-03    1250.80     0.0045         10.125              1.000
            2014-11-11   59888.21     0.1857         12.397              0.808
            2014-11-12   -3578.78    -0.0140          6.671              0.560
            2014-11-26  -29085.19    -0.0946         12.494              0.762
            ...               ...        ...            ...                ...
            2016-03-14   16220.57     0.0559          9.324              0.444
            2016-03-14  -25328.12    -0.1218          5.201              0.623
            2016-03-30  -29858.44    -0.0863          4.070              0.536
            2016-04-04    5373.76     0.0244          5.618              0.190
            2016-04-13  -28044.40    -0.1159          4.457              0.270
            2016-04-14  -18645.93    -0.0467          0.733              0.940
            2016-04-15  -32484.79    -0.1149          1.162              0.631
            2016-04-15  -32484.79    -0.1149          1.162              0.631
            2016-04-29     290.96     0.0007         -1.115              1.000
            2016-04-29     290.96     0.0007         -1.115              1.000

                        buy_wave_score3  buy_atr_std
            2014-09-24            0.218        0.226
            2014-10-24            0.449        0.146
            2014-10-29            0.329        0.538
            2014-10-29           -0.203        0.558
            2014-10-29           -0.173        0.063
            2014-10-29            0.329        0.538
            2014-11-03           -0.172       -0.002
            2014-11-11            0.001        0.171
            2014-11-12           -0.509        0.093
            2014-11-26            0.015        0.812
            ...                     ...          ...
            2016-03-14            0.948        0.126
            2016-03-14            0.788        0.663
            2016-03-30            0.702       -0.008
            2016-04-04            0.743        0.035
            2016-04-13            0.651        0.262
            2016-04-14            0.170        0.049
            2016-04-15            0.800        0.135
            2016-04-15            0.800        0.135
            2016-04-29            0.918        0.838
            2016-04-29            0.918        0.838
        """

        return mul_df


class EdgeMul(EdgeBase, BuyUmpMixin):
    """边裁单混特征类，UltronUmpEdgeBase子类，混入BuyUmpMixin，做为买入ump类"""

    def get_predict_col(self):
        """
        边裁单混特征keys：['buy_deg_ang21', 'buy_price_rank252', 'buy_wave_score3', 'buy_atr_std']
        :return: ['buy_deg_ang21', 'buy_price_rank252', 'buy_wave_score3', 'buy_atr_std']
        """

        return [
            FeatureDeg().get_feature_ump_keys(ump_cls=EdgeMul)[-1],
            FeaturePrice().get_feature_ump_keys(ump_cls=EdgeMul)[-1],
            FeatureWave().get_feature_ump_keys(ump_cls=EdgeMul)[-1],
            FeatureAtr().get_feature_ump_keys(ump_cls=EdgeMul)[-1]
        ]

    def get_fiter_class(self):
        """
        边裁单混特征返回的UltronMLPd子类：UltronUmpEdgeMul.UmpMulFiter
        :return: UltronUmpEdgeMul.UmpMulFiter
        """
        return MulFiter

    @classmethod
    def class_unique_id(cls):
        """
         具体ump类关键字唯一名称，类方法：return 'mul_edge'
         主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
         具体使用见UltronUmpManager中extend_ump_block方法
        """
        return 'mul_edge'



# -*- coding: utf-8 -*-

from sklearn.svm import NuSVR
from ultron.optimize.model.modelbase import create_model_base


class NvSVRModel(create_model_base('sklearn')):

    def __init__(self, features=None, fit_target=None, **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = NuSVR(**kwargs)


# -*- encoding:utf-8 -*-
"""买入因子类装饰器模块"""
from ultron.ump.core.fixes import six
from ultron.ump.technical.line import Line


class LeastPolyWrap(object):
    """示例做为买入因子策略装饰器封装least_valid_poly对大盘震荡大的情况下封锁交易"""

    def __call__(self, cls):
        """只做为买入因子策略类的装饰器"""

        if isinstance(cls, six.class_types):
            # 只做为类装饰器使用

            init_self = cls._init_self
            org_fit_day = cls.fit_day

            # fit_month不是必须实现的
            org_fit_month = getattr(cls, 'fit_month', None)

            def init_self_wrapped(*args, **kwargs):
                # 拿出被装饰的self对象
                warp_self = args[0]
                # 外部可以设置poly阀值，self.poly在fit_month中和每一个月大盘计算的poly比较，
                # 若是大盘的poly大于poly认为走势震荡
                warp_self.poly = kwargs.pop('poly', 2)
                # 是否封锁买入策略进行择时交易
                warp_self.lock = False
                # 调用原始的_init_self
                init_self(*args, **kwargs)

            def fit_day_wrapped(*args, **kwargs):
                # 拿出被装饰的self对象
                warp_self = args[0]
                if warp_self.lock:
                    # 如果封锁策略进行交易的情况下，策略不进行择时
                    return None
                return org_fit_day(*args, **kwargs)

            def fit_month_wrapped(*args, **kwargs):
                warp_self = args[0]
                today = args[1]
                # fit_month即在回测策略中每一个月执行一次的方法
                # 策略中拥有self.benchmark，即交易基准对象，AbuBenchmark实例对象，benchmark.kl_pd即对应的市场大盘走势
                benchmark_df = warp_self.benchmark.kl_pd
                # 拿出大盘的今天
                benchmark_today = benchmark_df[benchmark_df.date == today.date]
                if benchmark_today.empty:
                    return 0
                # 要拿大盘最近一个月的走势，准备切片的start，end
                end_key = int(benchmark_today.iloc[0].key)
                start_key = end_key - 20
                if start_key < 0:
                    return 0

                # 使用切片切出从今天开始向前20天的数据
                benchmark_month = benchmark_df.set_index(
                    'key').loc[start_key:end_key + 1].reset_index()

                # 通过大盘最近一个月的收盘价格做为参数构造AbuTLine对象
                benchmark_month_line = Line(benchmark_month.close,
                                            'benchmark month line')
                # 计算这个月最少需要几次拟合才能代表走势曲线
                least, _, _, _, _, _ = benchmark_month_line.create_least_valid_poly(
                )

                if least >= warp_self.poly:
                    # 如果最少的拟合次数大于阀值self.poly，说明走势成立，大盘非震荡走势，解锁交易
                    warp_self.lock = False
                else:
                    # 如果最少的拟合次数小于阀值self.poly，说明大盘处于震荡走势，封锁策略进行交易
                    warp_self.lock = True

                if org_fit_month is not None:
                    return org_fit_month(*args, **kwargs)

            cls._init_self = init_self_wrapped
            init_self_wrapped.__name__ = '_init_self'

            cls.fit_day = fit_day_wrapped
            fit_day_wrapped.__name__ = 'fit_day'

            cls.fit_month = fit_month_wrapped
            fit_month_wrapped.__name__ = 'fit_month'

            return cls
        else:
            raise TypeError('AbuLeastPolyWrap just for class warp')



# -*- encoding:utf-8 -*-
"""
    卖出择时示例因子：单日最大跌幅n倍atr止损
    做为单边止损因子使用，作为风险控制保护因子
"""

from ultron.ump.factor.sell.base import FactorSellBase, ESupportDirection
"""外部可通过如：abupy.fs.pre.g_default_pre_atr_n = 2.5来修改默认值"""
g_default_pre_atr_n = 1.5


class FactorPreAtrNStop(FactorSellBase):
    """示例单日最大跌幅n倍atr(止损)风险控制因子"""

    def _init_self(self, **kwargs):
        """kwargs中可选参数pre_atr_n: 单日最大跌幅止损的atr倍数"""

        self.pre_atr_n = g_default_pre_atr_n
        if 'pre_atr_n' in kwargs:
            # 设置下跌止损倍数
            self.pre_atr_n = kwargs['pre_atr_n']
            self.sell_type_extra = '{}:pre_atr={}'.format(
                self.__class__.__name__, self.pre_atr_n)

    def support_direction(self):
        """单日最大跌幅n倍atr(止损)因子支持两个方向"""
        return [
            ESupportDirection.DIRECTION_CAll.value,
            ESupportDirection.DIRECTION_PUT.value
        ]

    def fit_day(self, today, orders):
        """
        止损event：今天相比昨天的收益 * 买入时的期望方向 > today.atr21 * pre_atr_n
        :param today: 当前驱动的交易日金融时间序列数据
        :param orders: 买入择时策略中生成的订单序列
        :return:
        """

        for order in orders:
            if (today.pre_close - today.close
                ) * order.expect_direction > today.atr21 * self.pre_atr_n:
                # 只要今天的收盘价格比昨天收盘价格差大于一个差值就止损卖出, 亦可以使用其它计算差值方式
                self.sell_tomorrow(order)



# -*- encoding:utf-8 -*-
"""
    VWAP: Volume Weighted Average Price 成交量加权平均价

    非标准传统意义vwap计算，即非使用高频的分钟k线和量进行计算，只是套用概念计算
    日线级别的vwap
"""
from enum import Enum
import numpy as np
import pandas as pd
from ultron.ump.technical.line import Line


class EDayBarCalcType(Enum):
    """计算金融序列最终price使用的方法"""
    """高开低收的平均值"""
    OHLC_MEAN = 0
    """高开低收的中位数"""
    OHLC_MEDIAN = 1
    """高低的平均值"""
    HL_MEAN = 2
    """使用close价格"""
    CLOSE = 3


def calc_vwap(kl_pd, how=EDayBarCalcType.OHLC_MEAN):
    """
    非标准传统意义vwap计算，即非使用高频的分钟k线和量进行计算，只是套用概念计算
    日线级别的vwap，使用EDayBarCalcType确定日线级别上的最终点位vwap序列，即确定
    计算vwap使用的价格序列，需要注意成交量的数据准确度问题，且对异常成交量需要进行
    控制筛选
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param how: EDayBarCalcType，计算确定日线级别上的最终点位价格序列使用的方法
    :param show: 是否可视化，可视化使用UltronTLine.show接口
    :return: 返回UltronTLine对象
    """
    if how == EDayBarCalcType.OHLC_MEAN:
        # 日k级别上高开低收的平均
        price = np.mean([kl_pd.high, kl_pd.open, kl_pd.close, kl_pd.low],
                        axis=0)
    elif how == EDayBarCalcType.OHLC_MEDIAN:
        # 日k级别上高开低收的中位数
        price = np.median([kl_pd.high, kl_pd.open, kl_pd.close, kl_pd.low],
                          axis=0)
    elif how == EDayBarCalcType.HL_MEAN:
        # 日k级别上高低的平均价格
        price = np.mean([kl_pd.high, kl_pd.low], axis=0)
    elif how == EDayBarCalcType.CLOSE:
        # 日k级别上直接使用收close
        price = kl_pd.close
    else:
        raise TypeError('calc_vwap_std how is error! how= {}'.format(how))

    # 使用得到价格序列做基础计算vwap
    vwap = (kl_pd.volume * price).sum() / kl_pd.volume.sum()

    # 主要目的就是通过vwap构造line, 这里设置了price为line，使用mean=vwap，详阅UltronTLine
    line = Line(price, 'vwap std', mean=vwap)
    return line


# -*- encoding:utf-8 -*-
"""
    文件处理读取写入
"""

import logging, os, shutil
from contextlib import contextmanager

import functools
import pandas as pd

from ultron.kdutils.decorator import warnings_filter
# noinspection PyUnresolvedReferences
from ultron.ump.core.fixes import pickle, Pickler, Unpickler, as_bytes
"""HDF5_COMP_LEVEL：压缩级别：0-9，如果修改了压缩级别，需要删除之前的物理文件"""
HDF5_COMP_LEVEL = 4
""" HDF5_COMP_LIB: 使用的压缩库:'blosc', 'bzip2', 'lzo', 'zlib', 如果修改了压缩库，需要删除之前的物理文件"""
HDF5_COMP_LIB = 'blosc'
"""HDF5内部存贮依然会使用pickle，即python版本切换，本地文件会有协议冲突：使用所支持的最高协议进行dump"""
K_SET_PICKLE_HIGHEST_PROTOCOL = False
"""HDF5内部存贮依然会使用pickle，即python版本切换，本地文件会有协议冲突：python2, python3协议兼容模式，使用protocol=0"""
K_SET_PICKLE_ZERO_PROTOCOL = False


def ensure_dir(a_path):
    """
    确保a_path所在路径文件夹存在，如果a_path是文件将确保它上一级
    文件夹的存在
    :param a_path: str对象, 相对路径或者绝对路径
    """
    if os.path.isdir(a_path):
        a_dir = a_path
    else:
        a_dir = os.path.dirname(a_path)
    if not os.path.exists(a_dir):
        os.makedirs(a_dir)


def ensure_file(a_path):
    """
    确保a_path所在路径文件存在，首先会使用ensure_dir确保文件夹存在，
    确定a_path是file会确保文件存在
    :param a_path: str对象, 相对路径或者绝对路径
    :return:
    """
    ensure_dir(a_path)
    open(a_path, 'a+').close()


def file_exist(a_path):
    """
    a_path是否存在
    :param a_path: str对象, 相对路径或者绝对路径
    """
    return os.path.exists(a_path)


def copy_file(source, target_dir):
    """
    拷贝文件操作，支持文件夹拷贝操作
    :param source: 文件名或者文件夹，str对象, 相对路径或者绝对路径
    :param target_dir: 文件名或者文件夹，str对象, 相对路径或者绝对路径
    """

    if os.path.exists(source):
        logging.error('copy_file source={} not exists!'.format(source))
        return

    ensure_dir(target_dir)

    if os.path.isdir(source):
        shutil.copytree(source, target_dir)
    else:
        shutil.copy(source, target_dir)


def del_file(a_path):
    """
    删除文件操作，支持文件夹删除操作
    :param a_path: 文件名或者文件夹，str对象, 相对路径或者绝对路径
    """
    if not file_exist(a_path):
        return

    if os.path.isdir(a_path):
        shutil.rmtree(a_path)
    else:
        os.remove(a_path)


def load_pickle(file_name):
    """
    读取python序列化的本地文件
    :param file_name: 文件名，str对象, 相对路径或者绝对路径
    :return:
    """
    if not file_exist(file_name):
        logging.error('load_pickle file_name={} not exists!'.format(file_name))
        return None

    # TODO 根据文件大小，决定这里是否需要tip wait
    print('please wait! load_pickle....:', file_name)

    try:
        with open(file_name, "rb") as unpickler_file:
            unpickler = Unpickler(unpickler_file)
            ret = unpickler.load()
    except EOFError:
        print('unpickler file with EOFError, please check {} is 0kb!!!'.format(
            file_name))
        ret = {}

    return ret


def dump_pickle(input_obj, file_name, how='normal'):
    """
    存贮python序列化的本地文件
    :param input_obj: 需要进行序列化的对象
    :param file_name: 文件名，str对象, 相对路径或者绝对路径
    :param how: 序列化协议选择，默认normal不特殊处理，
                zero使用python2, python3协议兼容模式，使用protocol=0，
                high使用支持的最高协议
    """
    ensure_dir(file_name)

    print('please wait! dump_pickle....:', file_name)

    try:
        with open(file_name, "wb") as pick_file:
            if K_SET_PICKLE_HIGHEST_PROTOCOL or how == 'high':
                """使用所支持的最高协议进行dump"""
                pickle.dump(input_obj, pick_file, pickle.HIGHEST_PROTOCOL)
            elif K_SET_PICKLE_ZERO_PROTOCOL or how == 'zero':
                """python2, python3协议兼容模式，使用protocol=0"""
                pickle.dump(input_obj, pick_file, 0)
            else:
                pickler = Pickler(pick_file)
                pickler.dump(input_obj)
    except Exception as e:
        logging.exception(e)


"""hdf5批量处理时保存HDFStore对象，为避免反复open，close"""
__g_batch_h5s = None


def __start_batch_h5s(file_name, mode):
    """
    使用pd.HDFStore打开file_name对象保存在全局__g_batch_h5s中
    :param file_name: hdf5文件路径名
    :param mode: 打开hdf5文件模式。eg：w, r, a
    """
    global __g_batch_h5s
    __g_batch_h5s = pd.HDFStore(file_name,
                                mode,
                                complevel=HDF5_COMP_LEVEL,
                                complib=HDF5_COMP_LIB)


def __end_batch_h5s():
    """
    如果__g_batch_h5s中hdf5对象仍然是打开的，进行flush，后close
    """
    global __g_batch_h5s
    if __g_batch_h5s is not None and __g_batch_h5s.is_open:
        __g_batch_h5s.flush()
        __g_batch_h5s.close()
        __g_batch_h5s = None


def batch_h5s(h5_fn, mode='a'):
    """
    使用装饰器方式对hdf5操作进行批量处理，外部使用：
        eg： 详见SymbolPd.py
            @batch_h5s(h5s_fn)
            def _batch_save():
                for df_dict in df_dicts:
                    # 每一个df_dict是一个并行的序列返回的数据
                    for ind, (key_tuple, df) in enumerate(df_dict.values()):
                        # (key_tuple, df)是保存kl需要的数据, 迭代后直接使用save_kline_df
                        save_kline_df(df, *key_tuple)
                        if df is not None:
                            print("save kl {}_{}_{} {}/{}".format(key_tuple[0].value, key_tuple[1], key_tuple[2], ind,
                                                                  df.shape[0]))
                    # 完成一层循环一次，即批量保存完一个并行的序列返回的数据后，进行清屏
                    do_clear_output()
    :param h5_fn: hdf5文件路径名, 如果为None。即忽略整个批处理流程
    :param mode: 打开hdf5文件模式。eg：w, r, a
    :return:
    """

    def _batch_h5s(func):

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if h5_fn is not None:
                __start_batch_h5s(h5_fn, mode)
            ret = func(*args, **kwargs)
            if h5_fn is not None:
                __end_batch_h5s()
            return ret

        return wrapper

    return _batch_h5s


@contextmanager
def batch_ctx_h5s(h5_fn, mode='a'):
    """
    使用上下文管理器方式对hdf5操作进行批量处理，与装饰器模式batch_h5s
    功能相同，为外部不方便封装为具体操作函数时使用
        eg：
            with batch_ctx_h5s(h5s_fn):
                for df_dict in df_dicts:
                    # 每一个df_dict是一个并行的序列返回的数据
                    for ind, (key_tuple, df) in enumerate(df_dict.values()):
                        # (key_tuple, df)是保存kl需要的数据, 迭代后直接使用save_kline_df
                        save_kline_df(df, *key_tuple)
                        if df is not None:
                            print("save kl {}_{}_{} {}/{}".format(key_tuple[0].value, key_tuple[1], key_tuple[2], ind,
                                                                  df.shape[0]))
                    # 完成一层循环一次，即批量保存完一个并行的序列返回的数据后，进行清屏
                    do_clear_output()

    :param h5_fn: hdf5文件路径名, 如果为None。即忽略整个批处理流程
    :param mode: 打开hdf5文件模式。eg：w, r, a
    :return:
    """
    if h5_fn is not None:
        __start_batch_h5s(h5_fn, mode)

    yield

    if h5_fn is not None:
        __end_batch_h5s()


@warnings_filter
def dump_del_hdf5(file_name, dump_dict, del_array=None):
    """
    对hdf5进行删除和保存操作，del_array中的key先删除，后保存dump_dict
    字典中的数据
    :param file_name: hdf5文件路径名
    :param dump_dict: 需要保存到hdf5中的数据字典
    :param del_array: 需要从hdf5中删除的key序列
    """
    global __g_batch_h5s

    def do_dump_del_hdf5(h5s):
        """
        对hdf5进行删除和保存操作执行函数
        :param h5s: hdf5对象句柄
        :return:
        """
        if del_array is not None:
            # 先执行删除操作
            for del_key in del_array:
                if h5s.__contains__(del_key):
                    del h5s[del_key]

        for input_key in dump_dict:
            input_obj = dump_dict[input_key]
            h5s[input_key] = input_obj

    if __g_batch_h5s is None:
        # 如果批处理句柄没有打着，使用with pd.HDFStore先打开
        with pd.HDFStore(file_name,
                         'a',
                         complevel=HDF5_COMP_LEVEL,
                         complib=HDF5_COMP_LIB) as h5s_obj:
            do_dump_del_hdf5(h5s_obj)
    else:
        # 使用批处理句柄__g_batch_h5s操作
        do_dump_del_hdf5(__g_batch_h5s)


@warnings_filter
def dump_hdf5(file_name, input_obj, input_key):
    """
    对hdf5进行保存操作
    :param file_name: hdf5文件路径名
    :param input_obj: 保存的数据对象
    :param input_key: 保存的数据key
    """
    global __g_batch_h5s

    if __g_batch_h5s is None:
        # 如果批处理句柄没有打着，使用with pd.HDFStore先打开
        with pd.HDFStore(file_name,
                         'a',
                         complevel=HDF5_COMP_LEVEL,
                         complib=HDF5_COMP_LIB) as h5s:
            h5s[input_key] = input_obj
    else:
        # 使用批处理句柄__g_batch_h5s操作
        __g_batch_h5s[input_key] = input_obj


def del_hdf5(file_name, key):
    """
    对hdf5进行删除操作
    :param file_name: hdf5文件路径名
    :param key: 保存的数据key
    """
    if not file_exist(file_name):
        return

    if __g_batch_h5s is None:
        # 如果批处理句柄没有打着，使用with pd.HDFStore先打开
        with pd.HDFStore(file_name,
                         'a',
                         complevel=HDF5_COMP_LEVEL,
                         complib=HDF5_COMP_LIB) as h5s:
            if h5s.__contains__(key):
                del h5s[key]
    else:
        # 使用批处理句柄__g_batch_h5s操作
        if __g_batch_h5s.__contains__(key):
            del __g_batch_h5s[key]


def load_hdf5(file_name, key):
    """
    读取hdf5中的数据
    :param file_name: df5文件路径名
    :param key: 保存的数据key
    """
    global __g_batch_h5s

    if not file_exist(file_name):
        return None

    def _load_hdf5(h5s):
        load_obj = None
        if h5s.__contains__(key):
            try:
                load_obj = h5s[key]
            except (AttributeError, TypeError):
                # 'NoneType' attribute 'T' is just None
                # TypeError: 'len() of unsized object'
                # 低版本hdf5 bug导致的存贮异常情况的读取，忽略，计为正常损耗
                pass
        return load_obj

    if __g_batch_h5s is None:
        # 如果批处理句柄没有打着，使用with pd.HDFStore先打开
        with pd.HDFStore(file_name,
                         'a',
                         complevel=HDF5_COMP_LEVEL,
                         complib=HDF5_COMP_LIB) as h5s_obj:
            return _load_hdf5(h5s_obj)
    else:
        # 使用批处理句柄__g_batch_h5s操作
        return _load_hdf5(__g_batch_h5s)


def dump_df_csv(file_name, df):
    """
    将pd.DataFrame对象保存在csv中
    :param file_name: 保存csv的文件名称
    :param df: 需要保存的pd.DataFrame对象
    """
    if df is not None:
        # TODO 为效率，不应该在函数内部ensure_dir，确保使用dump_df_csv需要在外部ensure_dir
        ensure_dir(file_name)
        df.to_csv(file_name, columns=df.columns, index=True, encoding='utf-8')


def load_df_csv(file_name):
    """
    从csv文件中实例化pd.DataFrame对象
    :param file_name: 保存csv的文件名称
    :return: pd.DataFrame对象
    """
    if file_exist(file_name):
        return pd.read_csv(file_name, index_col=0)
    return None


def save_file(ct, file_name):
    """
    将内容ct保存文件
    :param ct: 内容str对象
    :param file_name: 保存的文件名称
    :return:
    """
    ensure_dir(file_name)
    with open(file_name, 'wb') as f:
        f.write(ct)



# -*- encoding:utf-8 -*-
"""
    手续费模块
"""
import numpy as np
import pandas as pd
from contextlib import contextmanager
from ultron.ump.core.fixes import partial
from ultron.utilities.logger import kd_logger
from ultron.ump.core.env import EMarketTargetType
from ultron.ump.market.futures import FuturesCn
from ultron.ump.trade.order import OrderMarket


class Commission(object):
    """交易手续费计算，记录，分析类，在Capital中实例化"""

    def __init__(self, commission_dict):
        """
        :param commission_dict: 代表用户自定义手续费计算dict对象，
                                key：buy_commission_func， 代表用户自定义买入计算费用方法
                                key：sell_commission_func，代表用户自定义卖出计算费用方法
        """
        self.commission_dict = commission_dict
        # 对象内部记录交易的pd.DataFrame对象，列设定
        self.df_columns = ['type', 'date', 'symbol', 'commission']
        # 构建手续费记录pd.DataFrame对象commission_df
        self.commission_df = pd.DataFrame(columns=self.df_columns)

    def __str__(self):
        """打印对象显示：如果有手续费记录，打印记录df，否则打印commission_df.info"""
        if self.commission_df.shape[0] == 0:
            return str(self.commission_df.info())
        return str(self.commission_df)

    __repr__ = __str__

    # noinspection PyMethodMayBeStatic
    def _commission_enter(self, a_order, market=None):
        """
        通过a_order对象进行交易对象市场类型转换，分配对应手续费计算方法
        :param a_order: 交易单对象AbuOrder实例
        :return:
        """

        # 如果使用env中统一设置，即不需要通过OrderMarket对单子查询市场，提高运行效率，详ABuMarket
        # 不同的市场不同的计算手续费方法
        if market == EMarketTargetType.E_MARKET_TARGET_US:
            # 美股
            calc_commission_func = calc_commission_us
        elif market == EMarketTargetType.E_MARKET_TARGET_CN:
            # a股
            calc_commission_func = calc_commission_cn
        elif market == EMarketTargetType.E_MARKET_TARGET_FUTURES_CN:
            # 期货
            calc_commission_func = partial(calc_commission_futures_cn,
                                           symbol_name=a_order.buy_symbol)
        else:
            raise TypeError('buy_stock market error!!!')
        return calc_commission_func

    @contextmanager
    def buy_commission_func(self, a_order):
        """
        外部用with as 返回的list中需要加入计算的最终结果，否则不进行内部交易费用记录
        :param a_order: 买单对象AbuOrder实例
        """
        market = OrderMarket(a_order).symbol_market

        if self.commission_dict is not None and 'buy_commission_func' in self.commission_dict:
            # 如果有自定义计算交易费的方法使用自定义的
            buy_func = self.commission_dict['buy_commission_func']
        else:
            buy_func = self._commission_enter(a_order, market)

        # 使用list因为是可变类型，需要将外面的结果带回来
        commission_list = list()
        yield buy_func, commission_list

        # 如果有外部有append，说明需要记录手续费，且执行计算成功
        if len(commission_list) == 1:
            commission = commission_list[0]
            # 将买单对象AbuOrder实例中的数据转换成交易记录需要的np.array对象
            record = np.array(
                ['buy', a_order.buy_date, a_order.buy_symbol,
                 commission]).reshape(1, 4)
            record_df = pd.DataFrame(record, columns=self.df_columns)
            self.commission_df = self.commission_df.append(record_df)
        else:
            kd_logger.info('buy_commission_func calc error')

    @contextmanager
    def sell_commission_func(self, a_order):
        """
        外部用with as 返回的list中需要加入计算的最终结果，否则不进行内部交易费用记录
        :param a_order: 卖单对象AbuOrder实例
        """
        market = OrderMarket(a_order).symbol_market
        if self.commission_dict is not None and 'sell_commission_func' in self.commission_dict:
            # 如果有自定义计算交易费的方法使用自定义的
            sell_func = self.commission_dict['sell_commission_func']
        else:
            sell_func = self._commission_enter(a_order, market)
        # 使用list因为是可变类型，需要将外面的结果带回来
        commission_list = list()

        yield sell_func, commission_list

        if len(commission_list) == 1:
            commission = commission_list[0]
            # 将卖单对象AbuOrder实例中的数据转换成交易记录需要的np.array对象
            record = np.array(
                ['sell', a_order.sell_date, a_order.buy_symbol,
                 commission]).reshape(1, 4)
            record_df = pd.DataFrame(record, columns=self.df_columns)
            self.commission_df = self.commission_df.append(record_df)
        else:
            kd_logger.info('sell_commission_func calc error!!!')


def calc_commission_us(trade_cnt, price):
    """
    美股计算交易费用：每股0.01，最低消费2.99
    :param trade_cnt: 交易的股数（int）
    :param price: 每股的价格（美元）（暂不使用，只是保持接口统一）
    :return: 计算结果手续费
    """
    # 每股手续费0.01
    commission = trade_cnt * 0.01
    if commission < 2.99:
        # 最低消费2.99
        commission = 2.99
    return commission


def calc_commission_cn(trade_cnt, price):
    """
    a股计算交易费用：印花税＋佣金： 印花税万3，佣金万2.5
    :param trade_cnt: 交易的股数（int）
    :param price: 每股的价格（人民币）
    :return: 计算结果手续费
    """
    cost = trade_cnt * price
    # 印花税万3，
    tax = cost * 0.0003
    # 佣金万2.5
    commission = cost * 0.00025
    # 佣金最低5
    commission = commission if commission > 5 else 5
    commission += tax
    return commission


# noinspection PyUnusedLocal
def calc_commission_futures_cn(trade_cnt, price, symbol_name):
    """
    期货计算交易费用：首先查询对应商品单位交易量（每手单位数量），以及每手手续费，再计算对应手续费
    :param trade_cnt: 交易的单位数量（int）
    :param price: 买入的价格（暂不使用，只是保持接口统一）
    :param symbol_name: 商品查询symbol
    :return: 计算结果手续费
    """
    min_unit = 10
    commission_unit = 10
    # 查询商品期货的对应df
    q_df = FuturesCn().query_symbol(symbol_name)
    if q_df is not None:
        # 每手单位数量
        min_unit = q_df.min_unit.values[0]
        # 每手交易手续费
        commission_unit = q_df.commission.values[0]
    commission = trade_cnt / min_unit * commission_unit
    return commission



# -*- encoding:utf-8 -*-
"""
    对各个依赖库不同版本，不同系统的规范进行统一以及问题修正模块
"""
import numbers
import warnings
import sklearn
import numpy as np


def _parse_version(version_string):
    """
    根据库中的__version__字段，转换为tuple，eg. '1.11.3'->(1, 11, 3)
    :param version_string: __version__字符串对象
    :return: tuple 对象
    """
    version = []
    for x in version_string.split('.'):
        try:
            version.append(int(x))
        except ValueError:
            version.append(x)
    return tuple(version)


skl_version = _parse_version(sklearn.__version__)

try:
    import six as six
except ImportError:
    warnings.warn("not installed six")

try:
    from functools import partial
except ImportError:
    warnings.warn("not installed functools partial")

try:
    from six.moves import xrange
except ImportError:
    warnings.warn("not installed functools partial")

try:
    from six.moves import zip, xrange, range, reduce, map, filter
except ImportError:
    warnings.warn("not installed functools partial")

try:
    from inspect import signature, Parameter
except ImportError:
    try:
        from funcsigs import signature, Parameter
    except ImportError:
        warnings.warn("not installed funcsigs")

try:
    from scipy.stats import rankdata
except ImportError:
    warnings.warn("not installed scipy")

try:
    # noinspection all
    from six.moves import cPickle as pickle
except ImportError:
    # noinspection all
    from six.moves import cPickle as pickle

if six.PY3:
    # noinspection PyProtectedMember
    Unpickler = pickle._Unpickler
    # noinspection PyProtectedMember
    Pickler = pickle._Pickler
else:
    Unpickler = pickle.Unpickler
    Pickler = pickle.Pickler

if six.PY3:

    def as_bytes(s):
        if isinstance(s, bytes):
            return s
        return s.encode('latin1')
else:
    as_bytes = str


# noinspection PyProtectedMember,PyUnresolvedReferences
def check_random_state(seed):
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, (numbers.Integral, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError('%r cannot be used to seed a numpy.random.RandomState'
                     ' instance' % seed)


try:
    skl_ver_big = skl_version >= (0, 18, 0)
except:
    skl_ver_big = True

if skl_ver_big:
    mean_squared_error_scorer = 'neg_mean_squared_error'
    mean_absolute_error_scorer = 'neg_mean_absolute_error'
    median_absolute_error_scorer = 'neg_median_absolute_error'
    log_loss = 'neg_log_loss'

    try:
        from sklearn.model_selection import train_test_split
        from sklearn.model_selection import learning_curve
        from sklearn.model_selection import cross_val_score
        from sklearn.model_selection import GridSearchCV
        # noinspection PyPep8Naming
        from sklearn.mixture import GaussianMixture as GMM

        class KFold(object):
            """
                sklearn将KFold移动到了model_selection，而且改变了用法，暂时不需要
                这么复杂的功能，将sklearn中关键代码简单实现，不from sklearn.model_selection import KFold
            """

            def __init__(self, n, n_folds=3, shuffle=False, random_state=None):
                if abs(n - int(n)) >= np.finfo('f').eps:
                    raise ValueError("n must be an integer")
                self.n = int(n)

                if abs(n_folds - int(n_folds)) >= np.finfo('f').eps:
                    raise ValueError("n_folds must be an integer")
                self.n_folds = n_folds = int(n_folds)

                if n_folds <= 1:
                    raise ValueError(
                        "k-fold cross validation requires at least one"
                        " train / test split by setting n_folds=2 or more,"
                        " got n_folds={0}.".format(n_folds))
                if n_folds > self.n:
                    raise ValueError(
                        ("Cannot have number of folds n_folds={0} greater"
                         " than the number of samples: {1}.").format(
                             n_folds, n))

                if not isinstance(shuffle, bool):
                    raise TypeError("shuffle must be True or False;"
                                    " got {0}".format(shuffle))
                self.shuffle = shuffle
                self.random_state = random_state

                self.idxs = np.arange(n)
                if shuffle:
                    rng = check_random_state(self.random_state)
                    rng.shuffle(self.idxs)

            def __iter__(self):
                ind = np.arange(self.n)
                for test_index in self._iter_test_masks():
                    train_index = np.logical_not(test_index)
                    train_index = ind[train_index]
                    test_index = ind[test_index]
                    yield train_index, test_index

            def _iter_test_masks(self):
                for test_index in self._iter_test_indices():
                    test_mask = self._empty_mask()
                    test_mask[test_index] = True
                    yield test_mask

            def _empty_mask(self):
                return np.zeros(self.n, dtype=np.bool)

            def _iter_test_indices(self):
                n = self.n
                n_folds = self.n_folds
                fold_sizes = (n // n_folds) * np.ones(n_folds, dtype=np.int)
                fold_sizes[:n % n_folds] += 1
                current = 0
                for fold_size in fold_sizes:
                    start, stop = current, current + fold_size
                    yield self.idxs[start:stop]
                    current = stop

            def __repr__(self):
                return '%s.%s(n=%i, n_folds=%i, shuffle=%s, random_state=%s)' % (
                    self.__class__.__module__,
                    self.__class__.__name__,
                    self.n,
                    self.n_folds,
                    self.shuffle,
                    self.random_state,
                )

            def __len__(self):
                return self.n_folds

    except ImportError:
        from sklearn.cross_validation import train_test_split
        from sklearn.cross_validation import KFold
        from sklearn.cross_validation import cross_val_score
        from sklearn.learning_curve import learning_curve
        from sklearn import cross_validation
        from sklearn.grid_search import GridSearchCV
        from sklearn.mixture import GMM
else:
    mean_squared_error_scorer = 'mean_squared_error'
    mean_absolute_error_scorer = 'mean_absolute_error'
    median_absolute_error_scorer = 'median_absolute_error'
    log_loss = 'log_loss'

    # noinspection PyUnresolvedReferences, PyDeprecation
    from sklearn.cross_validation import train_test_split
    # noinspection PyUnresolvedReferences, PyDeprecation
    from sklearn.cross_validation import KFold
    # noinspection PyUnresolvedReferences, PyDeprecation
    from sklearn.cross_validation import cross_val_score
    # noinspection PyUnresolvedReferences, PyDeprecation
    from sklearn.learning_curve import learning_curve
    # noinspection PyUnresolvedReferences, PyDeprecation
    from sklearn import cross_validation
    # noinspection PyUnresolvedReferences, PyDeprecation
    from sklearn.grid_search import GridSearchCV
    # noinspection PyUnresolvedReferences, PyDeprecation
    from sklearn.mixture import GMM


# -*- coding: utf-8 -*-

from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union
import numpy as np

from ultron.utilities.exceptions import PortfolioBuilderException

from .optimizers import (
    TargetVarianceOptimizer
)



def mean_variance_builder(er: np.ndarray,
                risk_model: Dict[str, Union[None, np.ndarray]],
                current_position: np.ndarray,
                vol_target: float,
                turn_over_target: float,
                lbound: Union[np.ndarray, float, None],
                ubound: Union[np.ndarray, float, None],
                ) -> Tuple[str, np.ndarray, np.ndarray]:
    cov = risk_model['cov']
    prob = TargetVarianceOptimizer(objective=-er,
            current_pos=current_position,
            lbound=lbound,ubound=ubound,
            target_vol=vol_target,
            target_turn_over=turn_over_target,
            cov=cov,
            )
    if prob.status() == 'optimal' or prob.status() == 'optimal_inaccurate':
        return prob.status(), prob.feval(), prob.x_value()
    else:
        raise PortfolioBuilderException(prob.status())



# -*- coding: utf-8 -*-

import numpy as np
from . utilities import group_mapping
from . utilities import transform
from . utilities import aggregate
from . utilities import array_index
from . utilities import simple_mean
from . utilities import simple_std
from . utilities import simple_sqrsum


def standardize(x: np.ndarray, groups: np.ndarray=None, ddof=1) -> np.ndarray:

    if groups is not None:
        groups = group_mapping(groups)
        mean_values = transform(groups, x, 'mean')
        std_values = transform(groups, x, 'std', ddof)

        return (x - mean_values) / np.maximum(std_values, 1e-8)
    else:
        return (x - simple_mean(x, axis=0)) / np.maximum(simple_std(x, axis=0, ddof=ddof), 1e-8)


def projection(x: np.ndarray, groups: np.ndarray=None, axis=1) -> np.ndarray:
    if groups is not None and axis == 0:
        groups = group_mapping(groups)
        projected = transform(groups, x, 'project')
        return projected
    else:
        return x / simple_sqrsum(x, axis=axis).reshape((-1, 1))


class Standardizer(object):

    def __init__(self, ddof: int=1):
        self.ddof = ddof
        self.mean = None
        self.std = None
        self.labels = None

    def fit(self, x: np.ndarray, groups: np.ndarray=None):
        if groups is not None:
            group_index = group_mapping(groups)
            self.mean = aggregate(group_index, x, 'mean')
            self.std = aggregate(group_index, x, 'std', self.ddof)
            self.labels = np.unique(groups)
        else:
            self.mean = simple_mean(x, axis=0)
            self.std = simple_std(x, axis=0, ddof=self.ddof)

    def transform(self, x: np.ndarray, groups: np.ndarray=None) -> np.ndarray:
        if groups is not None:
            index = array_index(self.labels, groups)
            return (x - self.mean[index]) / np.maximum(self.std[index], 1e-8)
        else:
            return (x - self.mean) / np.maximum(self.std, 1e-8)

    def __call__(self, x: np.ndarray, groups: np.ndarray=None) -> np.ndarray:
        return standardize(x, groups, self.ddof)



# -*- coding: utf-8 -*-

from ultron.strategy.fut_strategy import Strategy as FutStrategy
from ultron.strategy.stk_strategy import Strategy as StkStrategy

func_dict = {'fut': FutStrategy, 'stk': StkStrategy}


class StrategyEngine(object):

    @classmethod
    def create_class(cls, name):
        return func_dict[name] if name in func_dict else None


def create_params(**kwargs):

    ### 序列设置
    industry_effective = [] if 'industry_effective' not in kwargs else kwargs[
        'industry_effective']

    industry_invalid = [] if 'industry_invalid' not in kwargs else kwargs[
        'industry_invalid']

    riskstyle = [] if 'riskstyle' not in kwargs else kwargs['riskstyle']

    ### 通用参数设置
    weights_bandwidth = 0.1 if 'weights_bandwidth' not in kwargs else kwargs[
        'weights_bandwidth']

    method = 'fmv' if 'method' not in kwargs else kwargs['method']

    turn_over_target = 1.0 if 'turn_over_target' not in kwargs else kwargs[
        'turn_over_target']

    target_vol = 0.1 if 'target_vol' not in kwargs else kwargs['target_vol']

    cov_windows = 20 if 'cov_windows' not in kwargs else kwargs['cov_windows']

    cov_method = 'unshrunk' if 'cov_method' not in kwargs else kwargs[
        'cov_method']

    lbound = 0. if 'lbound' not in kwargs else kwargs['lbound']

    ubound = 0.04 if 'ubound' not in kwargs else kwargs['ubound']

    is_benchmark = 0 if 'is_benchmark' not in kwargs else kwargs['is_benchmark']

    ### benchmark 区间设置
    benchmark_boundary = 'relative' if 'benchmark_boundary' not in kwargs else kwargs[
        'benchmark_boundary']

    benchmark_lower = 1.001 if 'benchmark_lower' not in kwargs else kwargs[
        'benchmark_lower']
    benchmark_upper = 0.8 if 'benchmark_upper' not in kwargs else kwargs[
        'benchmark_upper']

    #### total 区间设置
    total_boundary = 'relative' if 'total_boundary' not in kwargs else kwargs[
        'total_boundary']
    total_lower = 0.001 if 'total_lower' not in kwargs else kwargs[
        'total_lower']
    total_upper = 0.01 if 'total_upper' not in kwargs else kwargs['total_upper']

    #### 上限行业区间设置
    effective_industry_boundary = 'absolute' if 'effective_industry_boundary' not in kwargs else kwargs[
        'effective_industry_boundary']
    effective_industry_lower = 0.0 if 'effective_industry_lower' not in kwargs else kwargs[
        'effective_industry_lower']
    effective_industry_upper = 0.20 if 'effective_industry_upper' not in kwargs else kwargs[
        'effective_industry_upper']

    #### 下限行业区间设置
    invalid_industry_boundary = 'absolute' if 'invalid_industry_boundary' not in kwargs else kwargs[
        'invalid_industry_boundary']
    invalid_industry_lower = 0.0 if 'invalid_industry_lower' not in kwargs else kwargs[
        'invalid_industry_lower']
    invalid_industry_upper = 0.20 if 'invalid_industry_upper' not in kwargs else kwargs[
        'invalid_industry_upper']

    riskstyle_boundary = 'absolute' if 'riskstyle_boundary' not in kwargs else kwargs[
        'riskstyle_boundary']
    riskstyle_lower = 0.0 if 'riskstyle_lower' not in kwargs else kwargs[
        'riskstyle_lower']
    riskstyle_upper = 0.20 if 'riskstyle_upper' not in kwargs else kwargs[
        'riskstyle_upper']

    neutralized_styles = None if 'neutralized_styles' not in kwargs else kwargs[
        'neutralized_styles']

    other_boundary = 'absolute' if 'other_boundary' not in kwargs else kwargs[
        'other_boundary']

    params = {}
    params['industry'] = {}
    params['riskstyle'] = {}
    ### 序列设置
    params['industry']['effective'] = industry_effective
    params['industry']['invalid'] = industry_invalid
    params['riskstyle'] = riskstyle

    ### 通用参数设置
    params['setting_params'] = {}
    params['setting_params']['weights_bandwidth'] = weights_bandwidth
    params['setting_params']['method'] = method
    params['setting_params']['turn_over_target'] = turn_over_target
    params['setting_params']['target_vol'] = target_vol
    params['setting_params']['cov_windows'] = cov_windows
    params['setting_params']['cov_method'] = cov_method
    params['setting_params']['lbound'] = lbound
    params['setting_params']['ubound'] = ubound
    params['setting_params']['is_benchmark'] = is_benchmark

    params['setting_params']['benchmark'] = {}
    params['setting_params']['total'] = {}
    params['setting_params']['other'] = {}

    ###
    params['setting_params']['other']['boundary'] = other_boundary
    params['setting_params']['other']['lower'] = 0.0
    params['setting_params']['other']['upper'] = 0.0

    # benchmark 区间设置
    params['setting_params']['benchmark']['boundary'] = benchmark_boundary
    params['setting_params']['benchmark']['lower'] = benchmark_lower
    params['setting_params']['benchmark']['upper'] = benchmark_upper

    # total 区间设置   条件6
    params['setting_params']['total']['boundary'] = total_boundary
    params['setting_params']['total']['lower'] = total_lower
    params['setting_params']['total']['upper'] = total_upper

    ### 此处考虑行业择时
    params['setting_params']['effective_industry'] = {}
    params['setting_params']['invalid_industry'] = {}

    #### effective_industry 上限行业区间设置
    params['setting_params']['effective_industry'][
        'boundary'] = effective_industry_boundary
    params['setting_params']['effective_industry'][
        'lower'] = effective_industry_lower
    params['setting_params']['effective_industry'][
        'upper'] = effective_industry_upper

    #### invalid_industry 下限行业区间设置
    params['setting_params']['invalid_industry'][
        'boundary'] = invalid_industry_boundary
    params['setting_params']['invalid_industry'][
        'lower'] = invalid_industry_lower
    params['setting_params']['invalid_industry'][
        'upper'] = invalid_industry_upper

    ### riskstyle 风格设置
    params['setting_params']['riskstyle'] = {}
    params['setting_params']['riskstyle']['boundary'] = riskstyle_boundary
    params['setting_params']['riskstyle']['lower'] = riskstyle_lower
    params['setting_params']['riskstyle']['upper'] = riskstyle_upper

    params['setting_params']['neutralized_styles'] = neutralized_styles
    return params


# -*- encoding:utf-8 -*-
"""
    择时并行多任务调度模块
"""
import pandas as pd
import numpy as np
from ultron.ump.core import env
from ultron.ump.core.process import EnvProcess
from ultron.kdutils.parallel import delayed, Parallel
from ultron.ump.market.market import split_k_market
from ultron.ump.trade import trade_execute
from ultron.ump.alpha.pick_time_execute import do_symbols_with_same_factors


class PickTimeMaster(object):
    """择时并行多任务调度类"""

    @classmethod
    def do_symbols_with_same_factors_process(
            cls,
            target_symbols,
            benchmark,
            buy_factors,
            sell_factors,
            capital,
            kl_pd_manager=None,
            n_process_kl=env.g_cpu_cnt *
        2 if env.g_is_mac_os else env.g_cpu_cnt,
            n_process_pick_time=env.g_cpu_cnt,
            show_progress=True):
        """
        将多个交易对象拆解为多份交易对象序列，多任务并行完成择时工作
        :param target_symbols: 多个择时交易对象序列
        :param benchmark: 交易基准对象，AbuBenchmark实例对象
        :param buy_factors: 买入因子序列
        :param sell_factors: 卖出因子序列
        :param capital: AbuCapital实例对象
        :param kl_pd_manager: 金融时间序列管理对象，AbuKLManager实例
        :param n_process_kl: 控制金融时间序列管理对象内部启动n_process_kl进程获取金融序列数据
        :param n_process_pick_time: 控制择时操作并行任务数量
        :param show_progress: 显示进度条，透传do_symbols_with_same_factors，默认True
        """

        if n_process_pick_time <= 0:
            # 因为下面要根据n_process_pick_time来split_k_market
            n_process_pick_time = env.g_cpu_cnt

        # 将target_symbols切割为n_process_pick_time个子序列，这样可以每个进程处理一个子序列
        process_symbols = split_k_market(n_process_pick_time,
                                         market_symbols=target_symbols)

        # 因为切割会有余数，所以将原始设置的进程数切换为分割好的个数, 即32 -> 33 16 -> 17
        n_process_pick_time = len(process_symbols)

        parallel = Parallel(
            1,  #n_process_pick_time,
            verbose=0,
            pre_dispatch='2*n_jobs')
        # 每个并行的进程通过do_symbols_with_same_factors及自己独立的子序列独立工作，注意kl_pd_manager装载了所有需要的数据
        p_env = EnvProcess()
        out = parallel(
            delayed(do_symbols_with_same_factors)(choice_symbols,
                                                  benchmark,
                                                  buy_factors,
                                                  sell_factors,
                                                  capital,
                                                  apply_capital=False,
                                                  kl_pd_manager=kl_pd_manager,
                                                  env=p_env,
                                                  show_progress=show_progress)
            for choice_symbols in process_symbols)
        orders_pd = None
        action_pd = None
        all_fit_symbols_cnt = 0
        for sub_out in out:
            # 将每个子序列进程的处理结果进行合并
            sub_orders_pd, sub_action_pd, sub_all_fit_symbols_cnt = sub_out
            orders_pd = sub_orders_pd if orders_pd is None else pd.concat(
                [orders_pd, sub_orders_pd])
            action_pd = sub_action_pd if action_pd is None else pd.concat(
                [action_pd, sub_action_pd])
            all_fit_symbols_cnt += sub_all_fit_symbols_cnt

        if orders_pd is not None and action_pd is not None:
            # 将合并后的结果按照时间及行为进行排序
            # noinspection PyUnresolvedReferences
            action_pd = action_pd.sort_values(['Date', 'action'])
            action_pd.index = np.arange(0, action_pd.shape[0])
            # noinspection PyUnresolvedReferences
            orders_pd = orders_pd.sort_values(['buy_date'])
            # 最后将所有的action作用在资金上，生成资金时序，及判断是否能买入
            trade_execute.apply_action_to_capital(capital,
                                                  action_pd,
                                                  kl_pd_manager,
                                                  show_progress=show_progress)

        return orders_pd, action_pd, all_fit_symbols_cnt


# -*- coding: utf-8 -*-
"""
Created on 2017-4-26

@author: cheng.li
"""

import numpy as np
from numpy import zeros
from numpy import zeros_like
from ultron.factor.data.utilities import groupby
from ultron.factor.data.utilities import set_value


def rank_builder(er: np.ndarray,
                 use_rank: int,
                 groups: np.ndarray = None,
                 masks: np.ndarray = None) -> np.ndarray:

    er = er.copy()

    if masks is not None:
        er[~masks] = -np.inf

    if er.ndim == 1 or (er.shape[0] == 1 or er.shape[1] == 1):
        # fast path methods for single column er
        neg_er = -er.flatten()
        length = len(neg_er)
        weights = zeros((length, 1))
        if groups is not None:
            index_diff, order = groupby(groups)
            start = 0
            for diff_loc in index_diff:
                current_index = order[start:diff_loc + 1]
                current_ordering = neg_er[current_index].argsort()
                current_ordering.shape = -1, 1
                set_value(weights, current_index[current_ordering[:use_rank]],
                          1.)
                start = diff_loc + 1
        else:
            ordering = neg_er.argsort()
            weights[ordering[:use_rank]] = 1.
        return weights.reshape(er.shape)
    else:
        neg_er = -er
        weights = zeros_like(er)

        if groups is not None:
            index_diff, order = groupby(groups)
            start = 0
            for diff_loc in index_diff:
                current_index = order[start:diff_loc + 1]
                current_ordering = neg_er[current_index].argsort(axis=0)
                set_value(weights, current_index[current_ordering[:use_rank]],
                          1)
                start = diff_loc + 1
        else:
            ordering = neg_er.argsort(axis=0)
            set_value(weights, ordering[:use_rank], 1.)
        return weights
